'''
    File: iterative_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 6, 2022
    Update: October 2, 2023
    
    Implementations of iterative voting methods.
'''
from pref_voting.voting_method import  *
from pref_voting.voting_method import _num_rank_last, _num_rank_first
from pref_voting.profiles import  _borda_score, _find_updated_profile
from pref_voting.margin_based_methods import split_cycle, minimax_scores
from pref_voting.c1_methods import top_cycle, gocha 
from pref_voting.rankings import Ranking
from pref_voting.social_welfare_function import swf
import copy
from itertools import permutations, product
import numpy as np
from pref_voting.voting_method_properties import ElectionTypes
from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies

def _instant_runoff_basic(profile,curr_cands = None):
    "The basic implementation of instant runoff"
    # need the total number of all candidates in a profile to check when all candidates have been removed   
    num_cands = profile.num_cands 
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data

    winners = [c for c in candidates 
               if _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    while len(winners) == 0:
        plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates 
                            if  not isin(cands_to_ignore,c)}  
        min_plurality_score = min(plurality_scores.values())
        lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                             if  plurality_scores[c] == min_plurality_score])

        # remove cands with lowest plurality score
        cands_to_ignore = np.concatenate((cands_to_ignore, lowest_first_place_votes), axis=None)
        if len(cands_to_ignore) == num_cands: # removed all of the candidates 
            winners = sorted(lowest_first_place_votes)
        else:
            winners = [c for c in candidates 
                       if not isin(cands_to_ignore,c) and _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]
     
    return sorted(winners)

def _instant_runoff_recursive(profile, curr_cands = None):
    "A recursive implementation of instant runoff"
    candidates = curr_cands if curr_cands is not None else profile.candidates
    cands_to_ignore = np.array([c for c in profile.candidates if c not in candidates])
    rs, rcounts = profile.rankings_counts # get all the ranking data
    plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates if not isin(cands_to_ignore,c)}  
    min_plurality_score = min(plurality_scores.values())
    lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                            if  plurality_scores[c] == min_plurality_score])

    if len(lowest_first_place_votes) == len(candidates):
        return sorted(lowest_first_place_votes)
    
    else:
        return _instant_runoff_recursive(profile, [c for c in candidates if c not in lowest_first_place_votes])


def _instant_runoff_for_truncated_linear_orders(profile, curr_cands = None, threshold = None, hide_warnings = True): 
    """
    Instant Runoff for Truncated Linear Orders.  Iteratively remove the candidates with the fewest number of first place votes, until there is a candidate with more than the threshold number of first-place votes. 
    If a threshold is not set, then it is strictly more than half of the non-empty ballots. 
    
    Args:
        profile (ProfileWithTies): An anonymous profile with no ties in the ballots (note that ProfileWithTies allows for truncated linear orders).
        threshold (int, float, optional): The threshold needed to win the election.  If it is not set, then it is strictly more than half of the remaining ballots.
        hide_warnings (bool, optional): Show or hide the warnings when more than one candidate is eliminated in a round.

    Returns: 
        A sorted list of candidates
    
    .. note:: This is the simultaneous version of instant runoff, not the parallel-universe tiebreaking version. It is intended to be run on profiles with large number of voters in which there is a very low probability of a tie in the fewest number of first place votes.  A warning is displayed when more than one candidate is eliminated. 

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles_with_ties import ProfileWithTies
        from pref_voting.iterative_methods import instant_runoff_for_truncated_linear_orders

        prof = ProfileWithTies([{0:1, 1:1},{0:1, 1:2, 2:3, 3:4}, {0:1, 1:3, 2:3}, {3:2}, {0:1}, {0:1}, {}, {}])
        prof.display()

        tprof, report = prof.truncate_overvotes()
        for r, new_r, count in report: 
            print(f"{r} --> {new_r}: {count}")
        tprof.display()
        instant_runoff_for_truncated_linear_orders.display(tprof)
    
    
    """
    
    assert all([not r.has_overvote() for r in profile.rankings]), "Instant Runoff is only defined when all the ballots are truncated linear orders."
    
    curr_cands = profile.candidates if curr_cands is None else curr_cands

    # we need to remove empty rankings during the algorithm, so make a copy of the profile
    prof2 = copy.deepcopy(profile) 
    
    _prof = prof2.remove_candidates([c for c in profile.candidates if c not in curr_cands])

    # remove the empty rankings
    _prof.remove_empty_rankings()
    
    threshold = threshold if threshold is not None else _prof.strict_maj_size()
    
    remaining_candidates = _prof.candidates
        
    pl_scores = _prof.plurality_scores()
    max_pl_score = max(pl_scores.values())
    
    while max_pl_score < threshold: 

        reduced_prof = _prof.remove_candidates([c for c in _prof.candidates if c not in remaining_candidates])
        
        # after removing the candidates, there might be some empty ballots.
        reduced_prof.remove_empty_rankings()

        pl_scores = reduced_prof.plurality_scores()
        min_pl_score = min(pl_scores.values())
            
        cands_to_remove = [c for c in pl_scores.keys() if pl_scores[c] == min_pl_score]

        if not hide_warnings and len(cands_to_remove) > 1: 
            print(f"Warning: multiple candidates removed in a round: {', '.join(map(str,cands_to_remove))}")
            
        if len(cands_to_remove) == len(reduced_prof.candidates): 
            # all remaining candidates have the same plurality score.
            break 
            
        # possibly update the threshold, so that it is a strict majority of the remaining ballots
        threshold = threshold if threshold is not None else reduced_prof.strict_maj_size()
        max_pl_score = max(pl_scores.values())

        remaining_candidates = [c for c in remaining_candidates if c not in cands_to_remove]


    reduced_prof = _prof.remove_candidates([c for c in _prof.candidates if c not in remaining_candidates])

    # after removing the candidates, there might be some empty ballots.
    reduced_prof.remove_empty_rankings()
        
    pl_scores = reduced_prof.plurality_scores()
    
    max_pl_score = max(pl_scores.values())
    
    return sorted([c for c in pl_scores.keys() if pl_scores[c] == max_pl_score])

@vm(name = "Instant Runoff",
    input_types=[ElectionTypes.PROFILE])
def instant_runoff(profile, curr_cands = None, algorithm = "basic", **kwargs):
    """
    If there is a majority winner then that candidate is the winner. If there is no majority winner, then remove all candidates that are ranked first by the fewest number of voters. Continue removing candidates with the fewest number first-place votes until there is a candidate with a majority of first place votes.  
    
    .. important::
        If there is more than one candidate with the fewest number of first-place votes, then *all* such candidates are removed from the profile. 
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        algorithm (str, optional): The algorithm to use.  Options are "basic" and "recursive".  The default is "basic".

    Returns: 
        A sorted list of candidates

    .. seealso::

        Instant Runoff is also known as "Ranked Choice", "Hare", and "Alternative Vote".

        Related functions:  :func:`pref_voting.iterative_methods.instant_runoff_tb`, :func:`pref_voting.iterative_methods.instant_runoff_put`, :func:`pref_voting.iterative_methods.instant_runoff_with_explanation`
   
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import instant_runoff, ranked_choice, alternative_vote, hare
        
        prof = Profile([[2, 1, 0], [0, 2, 1], [1, 2, 0]], [1, 2, 2])

        prof.display()
        instant_runoff.display(prof)
        ranked_choice.display(prof)
        alternative_vote.display(prof)
        hare.display(prof)

    """
    if isinstance(profile, Profile): 
        if algorithm == "basic":
            return _instant_runoff_basic(profile, curr_cands = curr_cands)
        
        elif algorithm == "recursive":
            return _instant_runoff_recursive(profile, curr_cands = curr_cands)
        
        else:
            raise ValueError("Algorithm must be either 'basic' or 'recursive'.")
    elif isinstance(profile, ProfileWithTies): 
        return _instant_runoff_for_truncated_linear_orders(profile, curr_cands = curr_cands, **kwargs)
# Create some aliases for instant runoff
instant_runoff.set_name("Hare")
hare = copy.deepcopy(instant_runoff)
hare.skip_registration = True
instant_runoff.set_name("Ranked Choice")
ranked_choice = copy.deepcopy(instant_runoff)
ranked_choice.skip_registration = True
instant_runoff.set_name("Alternative Vote")
alternative_vote = copy.deepcopy(instant_runoff)
alternative_vote.skip_registration = True


# reset the name Instant Runoff
instant_runoff.set_name("Instant Runoff")

@swf(name = "Instant Runoff Ranking")
def instant_runoff_ranking(profile, curr_cands = None):
    """Returns the reverse of the elimination order in the instant runoff voting process.

    Args:
        profile (Profile): An anonymous Profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A Ranking of the candidates.
    """
    
    candidates = curr_cands if curr_cands is not None else profile.candidates
    cands_to_ignore = np.array([c for c in profile.candidates if c not in candidates])
    rs, rcounts = profile.rankings_counts # get all the ranking data
    plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates if not isin(cands_to_ignore,c)}  
    min_plurality_score = min(plurality_scores.values())
    lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                            if  plurality_scores[c] == min_plurality_score])
    
    if len(lowest_first_place_votes) == len(candidates):
        full_tie = Ranking({c:0 for c in candidates})
        return full_tie
    
    else:
        rec_ranking = instant_runoff_ranking(profile, [c for c in candidates if c not in lowest_first_place_votes])
        max_rank = max(rec_ranking.ranks)
        rec_ranking_dict = rec_ranking.rmap
        ranking = Ranking({c: rec_ranking_dict[c] if not isin(lowest_first_place_votes,c) else max_rank+1 for c in candidates})

        return ranking

@vm(name = "Instant Runoff TB",
    input_types=[ElectionTypes.PROFILE])
def instant_runoff_tb(profile, curr_cands = None, tie_breaker = None):
    """Instant Runoff (``instant_runoff``) with tie breaking:  If there is  more than one candidate with the fewest number of first-place votes, then remove the candidate with lowest in the tie_breaker ranking from the profile.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        tie_breaker (List[int]): A list of the candidates in the profile to be used as a tiebreaker.

    Returns: 
        A sorted list of candidates
   
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_tb
        
        prof = Profile([[1, 2, 0], [2, 1, 0], [0, 1, 2]], [1, 1, 1])

        prof.display()
        print("no tiebreaker")
        instant_runoff.display(prof)
        print("tie_breaker = [0, 1, 2]")
        instant_runoff_tb.display(prof) 
        print("tie_breaker = [1, 2, 0]")
        instant_runoff_tb.display(prof, tie_breaker=[1, 2, 0])

    """

    # the tie_breaker is any linear order (i.e., list) of the candidates
    tb = tie_breaker if tie_breaker is not None else list(range(profile.num_cands))
    
    # need the total number of all candidates in a profile to check when all candidates have been removed   
    num_cands = profile.num_cands 
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    winners = [c for c in candidates 
               if _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    while len(winners) == 0:
        plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates if not isin(cands_to_ignore,c)}  
        min_plurality_score = min(plurality_scores.values())
        lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                             if plurality_scores[c] == min_plurality_score])
        
        cand_to_remove = lowest_first_place_votes[0]
        for c in lowest_first_place_votes[1:]: 
            if tb.index(c) < tb.index(cand_to_remove):
                cand_to_remove = c

        # remove cands with lowest plurality winners
        cands_to_ignore = np.concatenate((cands_to_ignore, cand_to_remove), axis=None)
        if len(cands_to_ignore) == num_cands: #all the candidates where removed
            winners = sorted(lowest_first_place_votes)
        else:
            winners = [c for c in candidates 
                       if not isin(cands_to_ignore,c) and _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]
     
    return sorted(winners)

@vm(name = "Instant Runoff PUT",
    input_types=[ElectionTypes.PROFILE])
def instant_runoff_put(profile, curr_cands = None):
    """
    Instant Runoff (:func:`instant_runoff`) with parallel universe tie-breaking (PUT), defined recursively: if there is a candidate with a strict majority of first-place votes, that candidate is the IRV-PUT winner; otherwise a candidate x is an IRV-PUT winner if there is some candidate y with a minimal number of first-place votes such that after removing y from the profile, x is an IRV-PUT winner.
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. warning:: 
        This will take a long time on profiles with many candidates having the same plurality scores

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_tb, instant_runoff_put
        
        prof = Profile([[1, 2, 0], [2, 1, 0], [0, 1, 2]], [1, 1, 1])

        prof.display()
        print("no tiebreaker")
        instant_runoff.display(prof)
        print("tie_breaker = [0, 1, 2]")
        instant_runoff_tb.display(prof, tie_breaker=[0, 1, 2]) 
        print("tie_breaker = [0, 2, 1]")
        instant_runoff_tb.display(prof, tie_breaker=[0, 2, 1])
        print("tie_breaker = [1, 0, 2]")
        instant_runoff_tb.display(prof, tie_breaker=[1, 0, 2])
        print("tie_breaker = [1, 2, 0]")
        instant_runoff_tb.display(prof, tie_breaker=[1, 2, 0])
        print("tie_breaker = [2, 0, 1]")
        instant_runoff_tb.display(prof, tie_breaker=[2, 0, 1])
        print("tie_breaker = [2, 1, 0]")
        instant_runoff_tb.display(prof, tie_breaker=[2, 1, 0])
        print()
        instant_runoff_put.display(prof)


    """
    
    candidates = profile.candidates if curr_cands is None else curr_cands 

    plurality_scores = profile.plurality_scores(candidates)

    strict_maj_size = profile.strict_maj_size()
    majority_winner = [cand for cand, score in plurality_scores.items() if score >= strict_maj_size]

    if len(majority_winner) > 0:
        return majority_winner
    
    original_num_cands = len(candidates)
    
    # immediately eliminate candidates with plurality score 0
    # this is safe, because every elimination order will eliminate all these candidates first (in some order)
    candidates = [cand for cand in candidates if plurality_scores[cand] > 0]
    if len(candidates) < original_num_cands:
        # if we removed some candidates, we need to update the plurality scores
        plurality_scores = profile.plurality_scores(candidates)

    # plurality losers
    worst_score = min(plurality_scores.values())
    cands_to_remove = [cand for cand, value in plurality_scores.items() if value == worst_score]
    
    winners = []
    for cand_to_remove in cands_to_remove:
        new_winners = instant_runoff_put(profile, curr_cands = [c for c in candidates if not c == cand_to_remove])
        winners = winners + new_winners
    
    return sorted(set(winners))


# Create some aliases for instant runoff
instant_runoff_put.set_name("Hare PUT")
hare_put = copy.deepcopy(instant_runoff_put)
hare_put.skip_registration = True
instant_runoff_put.set_name("Ranked Choice PUT")
ranked_choice_put = copy.deepcopy(instant_runoff_put)
ranked_choice_put.skip_registration = True

# reset the name Instant Runoff
instant_runoff_put.set_name("Instant Runoff PUT")


def instant_runoff_with_explanation(profile, curr_cands = None):
    """
    Instant Runoff with an explanation. In addition to the winner(s), return the order in which the candidates are eliminated as a list of lists.    

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

        A list describing the order in which candidates are eliminated

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_with_explanation


        prof = Profile([[2, 1, 0], [0, 2, 1], [1, 2, 0]], [1, 2, 2])
        prof.display()
        instant_runoff.display(prof)
        ws, exp = instant_runoff_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")

        prof = Profile([[1, 2, 0], [2, 1, 0], [0, 1, 2]], [1, 1, 1])
        prof.display()
        instant_runoff.display(prof)
        ws, exp = instant_runoff_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")

        prof = Profile([[2, 0, 1, 3], [2, 0, 3, 1], [3, 0, 1, 2],  [3, 2, 1, 0], [0, 2, 1, 3]], [1, 1, 1, 1, 1])
        prof.display()
        instant_runoff.display(prof)
        ws, exp = instant_runoff_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")

    """
    # need the total number of all candidates in a profile to check when all candidates have been removed   
    num_cands = profile.num_cands 
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    

    winners = [c for c in candidates 
               if _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]
    elims_list = list()

    while len(winners) == 0:
        plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates 
                            if  not isin(cands_to_ignore,c)}  
        min_plurality_score = min(plurality_scores.values())
        lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                             if  plurality_scores[c] == min_plurality_score])

        elims_list.append(list(lowest_first_place_votes))

        # remove cands with lowest plurality winners
        cands_to_ignore = np.concatenate((cands_to_ignore, lowest_first_place_votes), axis=None)
        if len(cands_to_ignore) == num_cands: # removed all of the candidates 
            winners = sorted(lowest_first_place_votes)
        else:
            winners = [c for c in candidates 
                       if not isin(cands_to_ignore,c) and _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]
     
    return sorted(winners), elims_list

@vm(name="Instant Runoff (Truncated Linear Orders)",
    input_types=[ElectionTypes.TRUNCATED_LINEAR_PROFILE])
def instant_runoff_for_truncated_linear_orders(profile, curr_cands = None, threshold = None, hide_warnings = True): 
    """
    Instant Runoff for Truncated Linear Orders.  Iteratively remove the candidates with the fewest number of first place votes, until there is a candidate with more than the threshold number of first-place votes. 
    If a threshold is not set, then it is strictly more than half of the non-empty ballots. 
    
    Args:
        profile (ProfileWithTies): An anonymous profile with no ties in the ballots (note that ProfileWithTies allows for truncated linear orders).
        threshold (int, float, optional): The threshold needed to win the election.  If it is not set, then it is strictly more than half of the remaining ballots.
        hide_warnings (bool, optional): Show or hide the warnings when more than one candidate is eliminated in a round.

    Returns: 
        A sorted list of candidates
    
    .. note:: This is the simultaneous version of instant runoff, not the parallel-universe tiebreaking version. It is intended to be run on profiles with large number of voters in which there is a very low probability of a tie in the fewest number of first place votes.  A warning is displayed when more than one candidate is eliminated. 

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles_with_ties import ProfileWithTies
        from pref_voting.iterative_methods import instant_runoff_for_truncated_linear_orders

        prof = ProfileWithTies([{0:1, 1:1},{0:1, 1:2, 2:3, 3:4}, {0:1, 1:3, 2:3}, {3:2}, {0:1}, {0:1}, {}, {}])
        prof.display()

        tprof, report = prof.truncate_overvotes()
        for r, new_r, count in report: 
            print(f"{r} --> {new_r}: {count}")
        tprof.display()
        instant_runoff_for_truncated_linear_orders.display(tprof)
    
    
    """
    
    assert all([not r.has_overvote() for r in profile.rankings]), "Instant Runoff is only defined when all the ballots are truncated linear orders."
    
    curr_cands = profile.candidates if curr_cands is None else curr_cands

    # we need to remove empty rankings during the algorithm, so make a copy of the profile
    prof2 = copy.deepcopy(profile) 
    
    _prof = prof2.remove_candidates([c for c in profile.candidates if c not in curr_cands])

    # remove the empty rankings
    _prof.remove_empty_rankings()
    
    threshold = threshold if threshold is not None else _prof.strict_maj_size()
    
    remaining_candidates = _prof.candidates
        
    pl_scores = _prof.plurality_scores()
    max_pl_score = max(pl_scores.values())
    
    while max_pl_score < threshold: 

        reduced_prof = _prof.remove_candidates([c for c in _prof.candidates if c not in remaining_candidates])
        
        # after removing the candidates, there might be some empty ballots.
        reduced_prof.remove_empty_rankings()

        pl_scores = reduced_prof.plurality_scores()
        min_pl_score = min(pl_scores.values())
            
        cands_to_remove = [c for c in pl_scores.keys() if pl_scores[c] == min_pl_score]

        if not hide_warnings and len(cands_to_remove) > 1: 
            print(f"Warning: multiple candidates removed in a round: {', '.join(map(str,cands_to_remove))}")
            
        if len(cands_to_remove) == len(reduced_prof.candidates): 
            # all remaining candidates have the same plurality score.
            break 
            
        # possibly update the threshold, so that it is a strict majority of the remaining ballots
        threshold = threshold if threshold is not None else reduced_prof.strict_maj_size()
        max_pl_score = max(pl_scores.values())

        remaining_candidates = [c for c in remaining_candidates if c not in cands_to_remove]


    reduced_prof = _prof.remove_candidates([c for c in _prof.candidates if c not in remaining_candidates])

    # after removing the candidates, there might be some empty ballots.
    reduced_prof.remove_empty_rankings()
        
    pl_scores = reduced_prof.plurality_scores()
    
    max_pl_score = max(pl_scores.values())
    
    return sorted([c for c in pl_scores.keys() if pl_scores[c] == max_pl_score])

def top_n_instant_runoff_for_truncated_linear_orders(
    profile, 
    n,
    curr_cands = None, 
    threshold = None, 
    hide_warnings = True): 
    """
    Returns the top n candidates according to the Instant Runoff method: Iteratively remove candidates until there are at most n candidates left.   Note that since there may be multiple candidates with the lowest plurality score, it may not be possible to reduce to exactly n candidates, in which case the function will return None.
    """
    
    assert all([not r.has_overvote() for r in profile.rankings]), "Instant Runoff is only defined when all the ballots are truncated linear orders."
    
    curr_cands = profile.candidates if curr_cands is None else curr_cands

    if len(curr_cands) <= n:
        return sorted(curr_cands)

    # we need to remove empty rankings during the algorithm, so make a copy of the profile
    prof2 = copy.deepcopy(profile) 
    
    _prof = prof2.remove_candidates([c for c in profile.candidates if c not in curr_cands])

    # remove the empty rankings
    _prof.remove_empty_rankings()
    
    remaining_candidates = _prof.candidates
        
    pl_scores = _prof.plurality_scores()
    
    while len(remaining_candidates) > n: 
        reduced_prof = _prof.remove_candidates([c for c in _prof.candidates if c not in remaining_candidates])
        
        # after removing the candidates, there might be some empty ballots.
        reduced_prof.remove_empty_rankings()

        pl_scores = reduced_prof.plurality_scores()
        min_pl_score = min(pl_scores.values())
        cands_to_remove = [c for c in pl_scores.keys() if pl_scores[c] == min_pl_score]

        if not hide_warnings and len(cands_to_remove) > 1: 
            print(f"Warning: multiple candidates removed in a round: {', '.join(map(str,cands_to_remove))}")
            
        if len(cands_to_remove) == len(reduced_prof.candidates): 
            # all remaining candidates have the same plurality score.
            remaining_candidates = reduced_prof.candidates
            break 

        remaining_candidates = [c for c in remaining_candidates if c not in cands_to_remove]

    if len(remaining_candidates) != n:
        if not hide_warnings:
            print(f"Warning: cannot reduce to exactly {n} candidates.")
        return None

    return sorted(remaining_candidates)


@vm(name="Bottom-Two-Runoff Instant Runoff",
    input_types=[ElectionTypes.PROFILE])
def bottom_two_runoff_instant_runoff(profile, curr_cands = None):
    """Find the two candidates with the lowest two plurality scores, remove the one who loses head-to-head to the other, and repeat until a single candidate remains. 
    
    If there is a tie for lowest or second lowest plurality score, consider all head-to-head matches between a candidate with lowest and a candidate with second lowest plurality score, and remove all the losers of the head-to-head matches, unless this would remove all candidates.

    .. note:: 
        BTR-IRV is a Condorcet consistent voting method, i.e., if a Condorcet winner exists, then BTR-IRV will elect the Condorcet winner. 

    .. seealso::

        Related functions:  :func:`pref_voting.iterative_methods.bottom_two_runoff_instant_runoff_put`

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates
    """
    candidates = profile.candidates if curr_cands is None else curr_cands 

    if len(candidates) == 1:
        return candidates

    plurality_scores = profile.plurality_scores(candidates)
    worst_score = min(plurality_scores.values())
    cands_with_lowest_plurality_score = [cand for cand, value in plurality_scores.items() if value == worst_score]

    if len(cands_with_lowest_plurality_score) > 1:
        cands_with_second_lowest_plurality_score = cands_with_lowest_plurality_score
    else:
        second_lowest_plurality_score = sorted(plurality_scores.values())[1]
        cands_with_second_lowest_plurality_score = [cand for cand, value in plurality_scores.items() if value == second_lowest_plurality_score]
    
    cands_to_remove = []

    for c1 in cands_with_lowest_plurality_score:
        for c2 in cands_with_second_lowest_plurality_score:
            if c1 != c2:
                if profile.margin(c1,c2) <= 0:
                    cands_to_remove.append(c1)
                else:
                    cands_to_remove.append(c2)
    
    if len(set(cands_to_remove)) == len(candidates):
        return candidates
    else:
        return bottom_two_runoff_instant_runoff(profile, [cand for cand in candidates if cand not in set(cands_to_remove)])

@vm(name="Bottom-Two-Runoff Instant Runoff PUT",
    input_types=[ElectionTypes.PROFILE])
def bottom_two_runoff_instant_runoff_put(profile, curr_cands = None):
    """Find the two candidates with the lowest two plurality scores, remove the one who loses head-to-head to the other, and repeat until a single candidate remains. Parallel-universe tiebreaking is used to break ties for lowest or second lowest plurality scores. 

    .. note:: 
        BTR-IRV is a Condorcet consistent voting method, i.e., if a Condorcet winner exists, then BTR-IRV will elect the Condorcet winner. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates
    """
    candidates = profile.candidates if curr_cands is None else curr_cands 

    if len(candidates) == 1:
        return candidates

    plurality_scores = profile.plurality_scores(candidates)
    worst_score = min(plurality_scores.values())
    cands_with_lowest_plurality_score = [cand for cand, value in plurality_scores.items() if value == worst_score]

    if len(cands_with_lowest_plurality_score) > 1:
        cands_with_second_lowest_plurality_score = cands_with_lowest_plurality_score
    else:
        second_lowest_plurality_score = sorted(plurality_scores.values())[1]
        cands_with_second_lowest_plurality_score = [cand for cand, value in plurality_scores.items() if value == second_lowest_plurality_score]
    
    winners = []

    for c1 in cands_with_lowest_plurality_score:
        for c2 in cands_with_second_lowest_plurality_score:
            if c1 != c2:
                if profile.margin(c1,c2) <= 0:
                    additional_winners = bottom_two_runoff_instant_runoff_put(profile, curr_cands = [c for c in candidates if not c == c1])
                else:
                    additional_winners = bottom_two_runoff_instant_runoff_put(profile, curr_cands = [c for c in candidates if not c == c2])
                
                winners = winners + additional_winners
    
    return sorted(set(winners))
    

@vm(name = "Plurality with Runoff PUT",
    input_types=[ElectionTypes.PROFILE])
def plurality_with_runoff_put(profile, curr_cands = None):
    """If there is a majority winner then that candidate is the Plurality with Runoff winner. Otherwise hold a runoff between the top two candidates: the candidate with the most first place votes and the candidate with the 2nd most first place votes (or perhaps tied for the most first place votes). In the case of multiple candidates tied for the most or 2nd most first place votes, use parallel-universe tiebreaking: a candidate is a Plurality with Runoff winner if it is a winner in some runoff as described. If the candidates are all tied for the most first place votes, then all candidates are winners.
        
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates
        
    .. note:: 
        Plurality with Runoff is the same as Instant Runoff when there are 3 candidates, but they can give different answers with 4 or more candidates. 

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import instant_runoff, plurality_with_runoff_put

        prof = Profile([[0, 1, 2, 3], [3, 1, 2, 0], [2, 0, 3, 1], [1, 2, 3, 0], [2, 3, 0, 1], [0, 3, 2, 1]], [2, 1, 2, 2, 1, 2])
        prof.display()
        instant_runoff.display(prof)
        plurality_with_runoff_put.display(prof)
    
    """    

    curr_cands = profile.candidates if curr_cands is None else curr_cands
    
    if len(curr_cands) == 1: 
        return list(curr_cands)
    
    plurality_scores = profile.plurality_scores(curr_cands = curr_cands)  

    max_plurality_score = max(plurality_scores.values())
    
    first = [c for c in curr_cands if plurality_scores[c] == max_plurality_score]

    second = list()
    if len(first) == 1:
        second_plurality_score = list(reversed(sorted(plurality_scores.values())))[1]
        second = [c for c in curr_cands if plurality_scores[c] == second_plurality_score]

    if len(second) > 0:
        all_runoff_pairs = product(first, second)
    else: 
        all_runoff_pairs = [(c1,c2) for c1,c2 in product(first, first) if c1 != c2]

    winners = list()
    for c1, c2 in all_runoff_pairs: 
        
        if profile.margin(c1,c2) > 0:
            winners.append(c1)
        elif profile.margin(c1,c2) < 0:
            winners.append(c2)
        elif profile.margin(c1,c2) == 0:
            winners.append(c1)
            winners.append(c2)
    
    return sorted(list(set(winners)))

def plurality_with_runoff_put_with_explanation(profile, curr_cands = None):
    """Plurality with Runoff with an explanation. In addition to the winner(s), return list of the pairs of candidate that move on to runoff round.    
  
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates
        
    """    

    curr_cands = profile.candidates if curr_cands is None else curr_cands
    
    if len(curr_cands) == 1: 
        return list(curr_cands)
    
    plurality_scores = profile.plurality_scores(curr_cands = curr_cands)  

    max_plurality_score = max(plurality_scores.values())
    
    first = [c for c in curr_cands if plurality_scores[c] == max_plurality_score]
    second = list()
    if len(first) == 1:
        second_plurality_score = list(reversed(sorted(plurality_scores.values())))[1]
        second = [c for c in curr_cands if plurality_scores[c] == second_plurality_score]

    if len(second) > 0:
        all_runoff_pairs = list(product(first, second))
    else: 
        all_runoff_pairs = [(c1,c2) for c1,c2 in product(first, first) if c1 != c2]

    winners = list()
    for c1, c2 in all_runoff_pairs: 
        
        if profile.margin(c1,c2) > 0:
            winners.append(c1)
        elif profile.margin(c1,c2) < 0:
            winners.append(c2)
        elif profile.margin(c1,c2) == 0:
            winners.append(c1)
            winners.append(c2)
    
    return sorted(list(set(winners))), list(all_runoff_pairs)

@vm(name = "Coombs",
    input_types=[ElectionTypes.PROFILE])
def coombs(profile, curr_cands = None):
    """If there is a majority winner then that candidate is the Coombs winner.   If there is no majority winner, then remove all candidates that are ranked last by the greatest number of voters.  Continue removing candidates with the most last-place votes until there is a candidate with a majority of first place votes.  
    
    .. important::
        If there is  more than one candidate with the largest number of last-place votes, then *all* such candidates are removed from the profile. 
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.iterative_methods.coombs_with_tb`, :func:`pref_voting.iterative_methods.coomb_put`, :func:`pref_voting.iterative_methods.coombs_with_explanation`
   
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import instant_runoff, coombs
        
        prof = Profile([[2, 1, 0], [0, 2, 1], [1, 2, 0]], [1, 2, 2])

        prof.display()
        coombs.display(prof)
        instant_runoff.display(prof)

    """   

    # need the total number of all candidates in a profile to check when all candidates have been removed   
    num_cands = profile.num_cands 
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    winners = [c for c in candidates 
               if _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    while len(winners) == 0:
        
        last_place_scores = {c: _num_rank_last(rs, rcounts, cands_to_ignore, c) for c in candidates 
                             if not isin(cands_to_ignore,c)}  
        max_last_place_score = max(last_place_scores.values())
        greatest_last_place_votes = np.array([c for c in last_place_scores.keys() 
                                              if  last_place_scores[c] == max_last_place_score])

        # remove candidates ranked last by the greatest number of voters
        cands_to_ignore = np.concatenate((cands_to_ignore, greatest_last_place_votes), axis=None)
        
        if len(cands_to_ignore) == num_cands: # removed all candidates 
            winners = list(greatest_last_place_votes)
        else:
            winners = [c for c in candidates 
                       if not isin(cands_to_ignore,c) and _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    return sorted(winners)

@vm(name = "Coombs TB",
    input_types=[ElectionTypes.PROFILE])
def coombs_tb(profile, curr_cands = None, tie_breaker=None):
    """
    Coombs with a fixed tie-breaking rule: The tie-breaking rule is any linear order (i.e., list) of the candidates.  The default rule is to order the candidates as follows: 0,....,num_cands-1.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        tie_breaker (List[int]): A list of the candidates in the profile to be used as a tiebreaker.

    Returns: 

        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import coombs, coombs_tb
        
        prof = Profile([[2, 0, 1], [0, 2, 1], [1, 0, 2], [2, 1, 0], [0, 1, 2]], [1, 1, 1, 2, 1])
        prof.display()
        print("no tiebreaker")
        coombs.display(prof)
        print("tie_breaker = [0, 1, 2]")
        coombs_tb.display(prof) 
        print("tie_breaker = [2, 1, 0]")
        coombs_tb.display(prof, tie_breaker=[2, 1, 0])

    """

    # the tie_breaker is any linear order (i.e., list) of the candidates
    tb = tie_breaker if tie_breaker is not None else list(range(profile.num_cands))

    num_cands = profile.num_cands   
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data

    winners = [c for c in candidates 
               if _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    while len(winners) == 0:
        
        last_place_scores = {c: _num_rank_last(rs, rcounts, cands_to_ignore, c) for c in candidates 
                             if not isin(cands_to_ignore,c)}  
        max_last_place_score = max(last_place_scores.values())
        greatest_last_place_votes = [c for c in last_place_scores.keys() if last_place_scores[c] == max_last_place_score]

        # select the candidate to remove using the tie-breaking rule (a linear order over the candidates)
        cand_to_remove = greatest_last_place_votes[0]
        for c in greatest_last_place_votes[1:]: 
            if tb.index(c) < tb.index(cand_to_remove):
                cand_to_remove = c
        
        # remove candidates ranked last by the greatest number of voters
        cands_to_ignore = np.concatenate((cands_to_ignore, np.array([cand_to_remove])), axis=None)
        
        if len(cands_to_ignore) == num_cands:
            winners = list(greatest_last_place_votes)
        else:
            winners = [c for c in candidates 
                       if not isin(cands_to_ignore,c) and _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    return sorted(winners)

@vm(name = "Coombs PUT",
    input_types=[ElectionTypes.PROFILE])
def coombs_put(profile, curr_cands = None):
    """Coombs with parallel universe tie-breaking (PUT), defined recursively: if there is a candidate with a strict majority of first-place votes, that candidate is the Coombs-PUT winner; otherwise a candidate x is a Coombs-PUT winner if there is some candidate y with a maximal number of last-place votes such that after removing y from the profile, x is a Coombs-PUT winner.
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. warning:: 
        This will take a long time on profiles with many candidates having the same number of last-place votes.

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import coombs, coombs_tb, coombs_put
        
        prof = Profile([[2, 0, 1], [1, 0, 2], [0, 1, 2]], [2, 1, 1])

        prof.display()
        print("no tiebreaker")
        coombs.display(prof)
        print("tie_breaker = [0, 1, 2]")
        coombs_tb.display(prof, tie_breaker=[0, 1, 2]) 
        print("tie_breaker = [0, 2, 1]")
        coombs_tb.display(prof, tie_breaker=[0, 2, 1])
        print("tie_breaker = [1, 0, 2]")
        coombs_tb.display(prof, tie_breaker=[1, 0, 2])
        print("tie_breaker = [1, 2, 0]")
        coombs_tb.display(prof, tie_breaker=[1, 2, 0])
        print("tie_breaker = [2, 0, 1]")
        coombs_tb.display(prof, tie_breaker=[2, 0, 1])
        print("tie_breaker = [2, 1, 0]")
        coombs_tb.display(prof, tie_breaker=[2, 1, 0])
        print()
        coombs_put.display(prof)
    """

    candidates = profile.candidates if curr_cands is None else curr_cands 
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    majority_winner = [cand for cand, value in profile.plurality_scores(candidates).items() if value >= strict_maj_size]

    if len(majority_winner) > 0:
        return majority_winner
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    last_place_scores = {c: _num_rank_last(rs, rcounts, cands_to_ignore, c) for c in candidates}  
    max_last_place_score = max(last_place_scores.values())
    cands_to_remove = [c for c in last_place_scores.keys() if last_place_scores[c] == max_last_place_score]

    winners = []
    for cand_to_remove in cands_to_remove:
        new_winners = coombs_put(profile, curr_cands = [c for c in candidates if not c == cand_to_remove])
        winners = winners + new_winners
    
    return sorted(set(winners))

def coombs_with_explanation(profile, curr_cands = None):
    """
    Coombs with an explanation. In addition to the winner(s), return the order in which the candidates are eliminated as a list of lists.    

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

        A list describing the order in which candidates are eliminated

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import coombs, coombs_with_explanation


        prof = Profile([[2, 1, 0], [0, 2, 1], [1, 2, 0]], [1, 2, 2])
        prof.display()
        coombs.display(prof)
        ws, exp = coombs_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")

        prof = Profile([[1, 0, 3, 2], [2, 3, 1, 0], [2, 0, 3, 1], [1, 2, 3, 0]], [1, 1, 1, 1])
        prof.display()
        coombs.display(prof)
        ws, exp = coombs_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")


    """
    num_cands = profile.num_cands 
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    strict_maj_size = profile.strict_maj_size()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    winners = [c for c in candidates 
               if _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    elims_list = list()
    while len(winners) == 0:
        
        last_place_scores = {c: _num_rank_last(rs, rcounts, cands_to_ignore, c) for c in candidates 
                             if not isin(cands_to_ignore,c)}  
        max_last_place_score = max(last_place_scores.values())
        greatest_last_place_votes = np.array([c for c in last_place_scores.keys() 
                                              if  last_place_scores[c] == max_last_place_score])

        elims_list.append(list(greatest_last_place_votes))
        # remove candidates ranked last by the greatest number of voters
        cands_to_ignore = np.concatenate((cands_to_ignore, greatest_last_place_votes), axis=None)
        
        if len(cands_to_ignore) == num_cands:
            winners = list(greatest_last_place_votes)
        else:
            winners = [c for c in candidates 
                       if not isin(cands_to_ignore,c) and _num_rank_first(rs, rcounts, cands_to_ignore, c) >= strict_maj_size]

    return sorted(winners), elims_list

@vm(name = "Baldwin",
    input_types=[ElectionTypes.PROFILE])
def baldwin(profile, curr_cands = None):
    """Iteratively remove all candidates with the lowest Borda score until a single candidate remains.  If, at any stage, all  candidates have the same Borda score,  then all (remaining) candidates are winners.

    .. note:: 
        Baldwin is a Condorcet consistent voting method, i.e., if a Condorcet winner exists, then Baldwin will elect the Condorcet winner. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.iterative_methods.baldwin_with_tb`, :func:`pref_voting.iterative_methods.baldwin`, :func:`pref_voting.iterative_methods.baldwin_with_explanation`
   
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import baldwin
        
        prof = Profile([[1, 0, 2, 3], [3, 1, 0, 2], [2, 0, 3, 1]], [2, 1, 1])

        prof.display()
        baldwin.display(prof)
    """
    all_num_cands = profile.num_cands  
    candidates = profile.candidates if curr_cands is None else curr_cands
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in candidates]), all_num_cands)
    num_cands = len(candidates)
    cands_to_ignore = np.empty(0) 
    borda_scores = {c: _borda_score(rankings, rcounts, len(candidates), c) for c in candidates}

    min_borda_score = min(list(borda_scores.values()))
    last_place_borda_scores = [c for c in candidates 
                               if c in borda_scores.keys() and borda_scores[c] == min_borda_score]
      
    cands_to_ignore = np.concatenate((cands_to_ignore, last_place_borda_scores), axis=None)
    
    winners = list()
    if cands_to_ignore.shape[0] ==  num_cands: # all candidates have lowest Borda score
        winners = sorted(last_place_borda_scores)
    else: # remove the candidates with lowest Borda score
        num_cands = len(candidates)
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
        
    while len(winners) == 0:
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) for c in candidates if not isin(cands_to_ignore, c)}
                
        min_borda_score = min(borda_scores.values())
        last_place_borda_scores = [c for c in borda_scores.keys() if borda_scores[c] == min_borda_score]
        
        cands_to_ignore = np.concatenate((cands_to_ignore, last_place_borda_scores), axis=None)
                
        if cands_to_ignore.shape[0] == num_cands: # removed all remaining candidates
            winners = sorted(last_place_borda_scores)
        elif num_cands - cands_to_ignore.shape[0] ==  1: # only one candidate remains
            winners = sorted([c for c in candidates if c not in cands_to_ignore])
        else: 
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
    return sorted(winners)

@vm(name = "Baldwin TB",
    input_types=[ElectionTypes.PROFILE])
def baldwin_tb(profile, curr_cands = None, tie_breaker=None):
    """
    Baldwin with a fixed tie-breaking rule: The tie-breaking rule is any linear order (i.e., list) of the candidates.  The default rule is to order the candidates as follows: 0,....,num_cands-1.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        tie_breaker (List[int]): A list of the candidates in the profile to be used as a tiebreaker.

    Returns: 

        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import baldwin, baldwin_tb
        
        prof = Profile([[0, 2, 1, 3], [1, 3, 0, 2], [2, 3, 0, 1]], [1, 1, 1])
        prof.display()
        print("no tiebreaker")
        baldwin.display(prof)
        print("tie_breaker = [0, 1, 2, 3]")
        baldwin_tb.display(prof) 
        print("tie_breaker = [2, 1, 0, 3]")
        baldwin_tb.display(prof, tie_breaker=[2, 1, 0, 3])

    """
    
    # the tie_breaker is any linear order (i.e., list) of the candidates
    tb = tie_breaker if tie_breaker is not None else list(range(profile.num_cands))

    if len(profile.candidates) <= 1:
        return sorted(profile.candidates)

    all_num_cands = profile.num_cands  
    candidates = profile.candidates if curr_cands is None else curr_cands
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in candidates]), all_num_cands)
    num_cands = len(candidates)
    cands_to_ignore = np.empty(0) 
    borda_scores = {c: _borda_score(rankings, rcounts, num_cands, c) for c in candidates}

    min_borda_score = min(list(borda_scores.values()))
    last_place_borda_scores = [c for c in candidates 
                               if c in borda_scores.keys() and borda_scores[c] == min_borda_score]
            
    cand_to_remove = last_place_borda_scores[0]
    for c in last_place_borda_scores[1:]: 
        if tb.index(c) < tb.index(cand_to_remove):
            cand_to_remove = c
    cands_to_ignore = np.concatenate((cands_to_ignore, np.array([cand_to_remove])), axis=None)
    
    winners = list()
    if cands_to_ignore.shape[0] ==  all_num_cands: # all candidates have lowest Borda score
        winners = sorted(last_place_borda_scores)
    else: # remove the candidates with lowest Borda score
        num_cands = len(candidates)
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
        
    while len(winners) == 0:
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) 
                        for c in candidates if not isin(cands_to_ignore, c)}
                
        min_borda_score = min(borda_scores.values())
        last_place_borda_scores = [c for c in borda_scores.keys() if borda_scores[c] == min_borda_score]
        
        # select the candidate to remove using the tie-breaking rule (a linear order over the candidates)
        cand_to_remove = last_place_borda_scores[0]
        for c in last_place_borda_scores[1:]: 
            if tb.index(c) < tb.index(cand_to_remove):
                cand_to_remove = c
        
        cands_to_ignore = np.concatenate((cands_to_ignore, np.array([cand_to_remove])), axis=None)
                
        if cands_to_ignore.shape[0] == num_cands: # removed all remaining candidates
            winners = sorted(last_place_borda_scores)
        elif num_cands - cands_to_ignore.shape[0] ==  1: # only one candidate remains
            winners = sorted([c for c in candidates if c not in cands_to_ignore])
        else: 
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
    return sorted(winners)

@vm(name = "Baldwin PUT",
    input_types=[ElectionTypes.PROFILE])
def baldwin_put(profile, curr_cands=None):
    """Baldwin with parallel universe tie-breaking (PUT), defined recursively: if there is a single candidate in the profile, that candidate wins; otherwise a candidate x is a Baldwin-PUT winner if there is some candidate y with a minimal Borda score such that after removing y from the profile, x is a Baldwin-PUT winner.
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import baldwin, baldwin_tb, baldwin_put
        
        prof = Profile([[1, 2, 0], [0, 1, 2], [2, 0, 1]], [1, 3, 2])

        prof.display()
        print("no tiebreaker")
        baldwin.display(prof)
        print("tie_breaker = [0, 1, 2]")
        baldwin_tb.display(prof, tie_breaker=[0, 1, 2]) 
        print("tie_breaker = [0, 2, 1]")
        baldwin_tb.display(prof, tie_breaker=[0, 2, 1])
        print("tie_breaker = [1, 0, 2]")
        baldwin_tb.display(prof, tie_breaker=[1, 0, 2])
        print("tie_breaker = [1, 2, 0]")
        baldwin_tb.display(prof, tie_breaker=[1, 2, 0])
        print("tie_breaker = [2, 0, 1]")
        baldwin_tb.display(prof, tie_breaker=[2, 0, 1])
        print("tie_breaker = [2, 1, 0]")
        baldwin_tb.display(prof, tie_breaker=[2, 1, 0])
        print()
        baldwin_put.display(prof)
    """

    num_original_cands = len(profile.candidates)
    candidates = profile.candidates if curr_cands is None else curr_cands 
    
    if len(candidates) == 1:
        return candidates
    
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])
    
    rankings, rcounts = profile.rankings_counts # get all the ranking data
    updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_original_cands)

    borda_scores = {c: _borda_score(updated_rankings, rcounts, num_original_cands - cands_to_ignore.shape[0], c) for c in candidates if not isin(cands_to_ignore, c)}
    min_borda_score = min(list(borda_scores.values()))
    
    cands_to_remove = [c for c in candidates if c in borda_scores.keys() and borda_scores[c] == min_borda_score]

    winners = []
    for cand_to_remove in cands_to_remove:
        new_winners = baldwin_put(profile, curr_cands = [c for c in candidates if not c == cand_to_remove])
        winners = winners + new_winners
    
    return sorted(set(winners))


def baldwin_with_explanation(profile, curr_cands = None):
    """Baldwin with an explanation. In addition to the winner(s), return the order in which the candidates are eliminated as a list of dictionaries specifying the Borda scores in the profile restricted to the candidates that have not been eliminated.    

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

        A list describing for each round, the candidates that are eliminated and the Borda scores of the remaining candidates (in the profile restricted to candidates that have not been eliminated)

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import baldwin, baldwin_with_explanation

        prof = Profile([[2, 1, 0], [0, 2, 1], [1, 2, 0]], [1, 2, 2])
        prof.display()
        baldwin.display(prof)
        ws, exp = baldwin_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")

        prof = Profile([[1, 0, 3, 2], [2, 3, 1, 0], [2, 0, 3, 1], [1, 2, 3, 0]], [1, 1, 1, 1])
        prof.display()
        baldwin.display(prof)
        ws, exp = baldwin_with_explanation(prof)
        print(f"winning set: {ws}")
        print(f"order of elimination: {exp}")


    """

    all_num_cands = profile.num_cands   
    candidates = profile.candidates if curr_cands is None else curr_cands
    elims_list = list()
        
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), all_num_cands)
    
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    borda_scores = {c: _borda_score(rankings, rcounts, len(candidates), c) for c in candidates}

    min_borda_score = min(list(borda_scores.values()))
    
    last_place_borda_scores = [c for c in candidates 
                               if c in borda_scores.keys() and borda_scores[c] == min_borda_score]
    elims_list.append([last_place_borda_scores, borda_scores])
    cands_to_ignore = np.concatenate((cands_to_ignore, np.array(last_place_borda_scores)), axis=None)
    
    winners = list()
    if cands_to_ignore.shape[0] ==  all_num_cands: # all candidates have lowest Borda score
        winners = sorted(last_place_borda_scores)
    else: # remove the candidates with lowest Borda score
        num_cands = len(candidates)
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
        
    while len(winners) == 0:
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) 
                        for c in candidates if not isin(cands_to_ignore, c)}
                
        min_borda_score = min(borda_scores.values())
        last_place_borda_scores = [c for c in borda_scores.keys() if borda_scores[c] == min_borda_score]
        elims_list.append([last_place_borda_scores, borda_scores])
        
        cands_to_ignore = np.concatenate((cands_to_ignore, np.array(last_place_borda_scores)), axis=None)
                
        if cands_to_ignore.shape[0] == all_num_cands: # removed all remaining candidates
            winners = sorted(last_place_borda_scores)
        elif num_cands - cands_to_ignore.shape[0] ==  1: # only one candidate remains
            winners = sorted([c for c in candidates if c not in cands_to_ignore])
        else: 
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
    return sorted(winners), elims_list

@vm(name = "Strict Nanson",
    input_types=[ElectionTypes.PROFILE])
def strict_nanson(profile, curr_cands = None):
    """Iteratively remove all candidates with the  Borda score strictly below the average Borda score until one candidate remains.  If, at any stage, all  candidates have the same Borda score, then all (remaining) candidates are winners.

    .. note:: 
        
        Strict Nanson is a Condorcet consistent voting method, i.e., if a Condorcet winner exists, then Strict Nanson will elect the Condorcet winner. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.iterative_methods.strict_nanson_with_explanation`, :func:`pref_voting.iterative_methods.weak_nanson`
   
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import strict_nanson
        
        prof = Profile([[2, 1, 0, 3], [0, 2, 1, 3], [1, 3, 0, 2], [0, 3, 2, 1]], [2, 1, 1, 1])

        prof.display()
        strict_nanson.display(prof)
    """
    
    all_num_cands = profile.num_cands   
    candidates = profile.candidates if curr_cands is None else curr_cands
        
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), all_num_cands)
    cands_to_ignore = np.empty(0)

    borda_scores = {c: _borda_score(rankings, rcounts, len(candidates), c) for c in candidates}
    
    avg_borda_score = np.mean(list(borda_scores.values()))
    below_borda_avg_candidates = np.array([c for c in borda_scores.keys() if borda_scores[c] < avg_borda_score])
    
    cands_to_ignore = np.concatenate((cands_to_ignore, below_borda_avg_candidates), axis=None)
    winners = list()
    if cands_to_ignore.shape[0] == all_num_cands:  # all candidates have same Borda score
        winners = sorted(candidates)
    else: 
        num_cands = len(candidates)
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
    while len(winners) == 0: 
        
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) 
                        for c in candidates if not isin(cands_to_ignore, c)}
        
        avg_borda_scores = np.mean(list(borda_scores.values()))
    
        below_borda_avg_candidates = np.array([c for c in borda_scores.keys() 
                                               if borda_scores[c] < avg_borda_scores])
        
        cands_to_ignore = np.concatenate((cands_to_ignore, below_borda_avg_candidates), axis=None)
                
        if (below_borda_avg_candidates.shape[0] == 0) or ((all_num_cands - cands_to_ignore.shape[0]) == 1):
            winners = sorted([c for c in candidates if c not in cands_to_ignore])
        else:
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
            
    return winners


def strict_nanson_with_explanation(profile, curr_cands = None):
    """Strict Nanson with an explanation. In addition to the winner(s), return the order in which the candidates are eliminated as a list of dictionaries specifying the Borda scores in the profile restricted to the candidates that have not been eliminated and the average Borda score.    
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

        A list describing for each round, the candidates that are eliminated and the Borda scores of the remaining candidates (in the profile restricted to candidates that have not been eliminated)
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import strict_nanson_with_explanation
        
        prof = Profile([[2, 1, 0, 3], [0, 2, 1, 3], [1, 3, 0, 2], [0, 3, 2, 1]], [2, 1, 1, 1])

        prof.display()
        print(strict_nanson_with_explanation(prof))
    """

    all_num_cands = profile.num_cands   
    candidates = profile.candidates if curr_cands is None else curr_cands
    
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), all_num_cands)
    cands_to_ignore = np.empty(0)
    elim_list = list()
    
    borda_scores = {c: _borda_score(rankings, rcounts, len(candidates), c) for c in candidates}
    
    avg_borda_score = np.mean(list(borda_scores.values()))
    below_borda_avg_candidates = [c for c in borda_scores.keys() if borda_scores[c] < avg_borda_score]
    
    cands_to_ignore = np.concatenate((cands_to_ignore, np.array(below_borda_avg_candidates)), axis=None)
    winners = list()
    if cands_to_ignore.shape[0] == all_num_cands:  # all candidates have same Borda score
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
        winners = sorted(candidates)
    else: 
        num_cands = len(candidates)
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
    while len(winners) == 0: 
        
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) 
                        for c in candidates if not isin(cands_to_ignore, c)}
        
        avg_borda_score = np.mean(list(borda_scores.values()))
    
        below_borda_avg_candidates = [c for c in borda_scores.keys() 
                                      if borda_scores[c] < avg_borda_score]
        
        cands_to_ignore = np.concatenate((cands_to_ignore, np.array(below_borda_avg_candidates)), axis=None)
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
                
        if (len(below_borda_avg_candidates) == 0) or ((all_num_cands - cands_to_ignore.shape[0]) == 1):
            winners = sorted([c for c in candidates if c not in cands_to_ignore])
        else:
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
            
    return winners, elim_list

@vm(name = "Weak Nanson",
    input_types=[ElectionTypes.PROFILE])
def weak_nanson(profile, curr_cands = None):
    """Iteratively remove all candidates with Borda score less than or equal the average Borda score until one candidate remains.  If, at any stage, all  candidates have the same Borda score, then all (remaining) candidates are winners.

    .. note:: 

        Weak Nanson is a Condorcet consistent voting method, i.e.,  if a Condorcet winner exists, then Weak Nanson will elect the Condorcet winner. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :func:`pref_voting.iterative_methods.weak_nanson_with_explanation`, :func:`pref_voting.iterative_methods.strict_nanson`
   
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import weak_nanson
        
        prof = Profile([[2, 1, 0, 3], [0, 2, 1, 3], [1, 3, 0, 2], [0, 3, 2, 1]], [2, 1, 1, 1])

        prof.display()
        weak_nanson.display(prof)

    """

    all_num_cands = profile.num_cands   
    candidates = profile.candidates if curr_cands is None else curr_cands
        
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), all_num_cands)
    
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    borda_scores = {c: _borda_score(rankings, rcounts, len(candidates), c) for c in candidates}

    avg_borda_score = np.mean(list(borda_scores.values()))

    below_borda_avg_candidates = np.array([c for c in borda_scores.keys() if borda_scores[c] <= avg_borda_score])
    
    cands_to_ignore = np.concatenate((cands_to_ignore, below_borda_avg_candidates), axis=None)
    
    winners = list()
    if cands_to_ignore.shape[0] == all_num_cands:  # all candidates have same Borda score
        winners = sorted(candidates)
    elif all_num_cands - cands_to_ignore.shape[0]  == 1: # one candidate remains
        winners = [c for c in candidates if not isin(cands_to_ignore, c)]
    else: 
        num_cands = len(candidates)
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
        
    while len(winners) == 0: 
        
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) 
                        for c in candidates if not isin(cands_to_ignore, c)}
        

        avg_borda_score = np.mean(list(borda_scores.values()))

        below_borda_avg_candidates = np.array([c for c in borda_scores.keys() 
                                               if borda_scores[c] <= avg_borda_score])
        
        cands_to_ignore = np.concatenate((cands_to_ignore, below_borda_avg_candidates), axis=None)
        
        if cands_to_ignore.shape[0] == all_num_cands:  # all remaining candidates have been removed
            winners = sorted(below_borda_avg_candidates)
        elif all_num_cands - cands_to_ignore.shape[0]  == 1: # one candidate remains
            winners = [c for c in candidates if not isin(cands_to_ignore, c)]
        else:
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
            
    return winners


def weak_nanson_with_explanation(profile, curr_cands = None):
    """
    Weak Nanson with an explanation. In addition to the winner(s), return the order in which the candidates are eliminated as a list of dictionaries specifying the Borda scores in the profile restricted to the candidates that have not been eliminated and the average Borda score.    
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

        A list describing for each round, the candidates that are eliminated and the Borda scores of the remaining candidates (in the profile restricted to candidates that have not been eliminated)
    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import weak_nanson_with_explanation
        
        prof = Profile([[2, 1, 0, 3], [0, 2, 1, 3], [1, 3, 0, 2], [0, 3, 2, 1]], [2, 1, 1, 1])

        prof.display()
        print(weak_nanson_with_explanation(prof))
        
    """
    all_num_cands = profile.num_cands   
    candidates = profile.candidates if curr_cands is None else curr_cands
        
    rcounts = profile._rcounts # get all the ranking data
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), all_num_cands)
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])
    elim_list = list()
    
    borda_scores = {c: _borda_score(rankings, rcounts, len(candidates), c) for c in candidates}
    
    avg_borda_score = np.mean(list(borda_scores.values()))
    below_borda_avg_candidates = [c for c in borda_scores.keys() 
                                  if borda_scores[c] <= avg_borda_score]
    
    cands_to_ignore = np.concatenate((cands_to_ignore, np.array(below_borda_avg_candidates)), axis=None)
    winners = list()
    if cands_to_ignore.shape[0] == all_num_cands:  # all candidates have same Borda score
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
        winners = sorted(candidates)
    elif all_num_cands - cands_to_ignore.shape[0]  == 1: # one candidate remains
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
        winners = [c for c in candidates if not isin(cands_to_ignore, c)]
    else: 
        num_cands = len(candidates)
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
        updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
        

    while len(winners) == 0: 
        
        borda_scores = {c: _borda_score(updated_rankings, rcounts, num_cands - cands_to_ignore.shape[0], c) 
                        for c in candidates if not isin(cands_to_ignore, c)}
        
        avg_borda_score = np.mean(list(borda_scores.values()))
    
        below_borda_avg_candidates = [c for c in borda_scores.keys() 
                                      if borda_scores[c] <= avg_borda_score]
        
        cands_to_ignore = np.concatenate((cands_to_ignore, np.array(below_borda_avg_candidates)), axis=None)
        elim_list.append({"avg_borda_score": avg_borda_score, 
                          "elim_cands": below_borda_avg_candidates,
                          "borda_scores": borda_scores})
                
        if cands_to_ignore.shape[0] == all_num_cands:  # all remaining candidates have been removed
            winners = sorted(below_borda_avg_candidates)
        elif all_num_cands - cands_to_ignore.shape[0]  == 1: # one candidate remains
            winners = [c for c in candidates if not isin(cands_to_ignore, c)]
        else:
            updated_rankings = _find_updated_profile(rankings, cands_to_ignore, num_cands)
            
    return winners, elim_list


@vm(name = "Iterated Removal Condorcet Loser",
    input_types=[ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MAJORITY_GRAPH, ElectionTypes.MARGIN_GRAPH])
def iterated_removal_cl(edata, curr_cands = None):
    """
    Iteratively remove candidates that are Condorcet losers until there are no Condorcet losers.   A candidate :math:`c` is a **Condorcet loser** when every other candidate is majority preferred to :math:`c`. 

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `condorcet_loser` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        :meth:`pref_voting.profiles.Profile.condorcet_loser`,  :meth:`pref_voting.profiles_with_ties.ProfileWithTies.condorcet_loser`, :meth:`pref_voting.weighted_majority_graphs.MajorityGraph.condorcet_loser`

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.profiles_with_ties import ProfileWithTies
        from pref_voting.iterative_methods import iterated_removal_cl
        
        prof = Profile([[2, 1, 3, 0], [2, 1, 0, 3], [3, 1, 2, 0], [1, 2, 3, 0]], [1, 1, 1, 1])

        prof.display()
        iterated_removal_cl.display(prof)
        iterated_removal_cl.display(prof.majority_graph())
        iterated_removal_cl.display(prof.margin_graph())

        prof2 = ProfileWithTies([{2:1, 1:1, 3:2, 0:3}, {2:1, 1:2, 0:3, 3:4}, {3:1, 1:2, 2:3, 0:4}, {1:1, 2:2, 3:3, 0:4}], [1, 1, 1, 1])

        prof2.display()
        iterated_removal_cl.display(prof2)

    """

    condorcet_loser = edata.condorcet_loser(curr_cands = curr_cands)  
    
    remaining_cands = edata.candidates if curr_cands is None else curr_cands
    
    while len(remaining_cands) > 1 and  condorcet_loser is not None:    
        remaining_cands = [c for c in remaining_cands if c not in [condorcet_loser]]
        condorcet_loser = edata.condorcet_loser(curr_cands = remaining_cands)
            
    return sorted(remaining_cands)


def iterated_removal_cl_with_explanation(edata, curr_cands = None):
    """
    Iterated Removal Condorcet Loser with an explanation. In addition to the winner(s), return the order of elimination, where each candidate in the list is a Condorcet loser in the profile (restricted to the remaining candidates). 

    Args:
        edata (Profile, ProfileWithTies, MajorityGraph, MarginGraph): Any election data that has a `condorcet_loser` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.iterative_methods import iterated_removal_cl_with_explanation
        
        prof = Profile([[2, 1, 3, 0], [2, 1, 0, 3], [3, 1, 2, 0], [1, 2, 3, 0]], [1, 1, 1, 1])

        prof.display()
        ws, exp = iterated_removal_cl_with_explanation(prof)
        print(f"The winning set is {ws}")
        print(f"The order of elimination is {exp}")
    """

    elim_list = list()
    condorcet_loser = edata.condorcet_loser(curr_cands = curr_cands)  
    
    remaining_cands = edata.candidates if curr_cands is None else curr_cands
    
    while len(remaining_cands) > 1 and  condorcet_loser is not None: 
        elim_list.append(condorcet_loser)   
        remaining_cands = [c for c in remaining_cands if c not in [condorcet_loser]]
        condorcet_loser = edata.condorcet_loser(curr_cands = remaining_cands)
            
    return sorted(remaining_cands), elim_list

def _remove_worst_losers(edata,curr_cands,score_method):
    m_scores = minimax_scores(edata,curr_cands,score_method)
    worst_m_score = min([m_scores[c] for c in curr_cands])
    worst_losers = [c for c in curr_cands if m_scores[c] == worst_m_score]
    if len(worst_losers) == len(curr_cands):
        return curr_cands
    else:
        return [c for c in curr_cands if c not in worst_losers]

@vm(name = "Raynaud",
    input_types=[ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MARGIN_GRAPH])
def raynaud(edata, curr_cands=None, score_method = "margins"):
    """Iteratively remove the candidate(s) whose worst loss is biggest, unless all candidates have the same worst loss. See https://electowiki.org/wiki/Raynaud.
    
    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any election data that has a `margin` method. 
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``.
        score_method (str, optional): Options include "margins" (the default), "winning" assigns to each candidate :math:`c` the maximum support of a candidate majority preferred to :math:`c`,  and "pairwise_opposition" assigns to each candidate :math:`c` the maximum support of any candidate over :math:`c`.   These scores only lead to different results on non-linear profiles. 

    Returns: 
        A sorted list of candidates. 
    """
    candidates = edata.candidates if curr_cands is None else curr_cands
    new_cands = _remove_worst_losers(edata,candidates,score_method)
    while not new_cands == candidates:
        candidates = new_cands
        new_cands = _remove_worst_losers(edata,candidates,score_method)
    return sorted(candidates)

@vm(name = "Benham",
    input_types=[ElectionTypes.PROFILE])
def benham(profile, curr_cands = None):
    """
    As long as the profile has no Condorcet winner, eliminate the candidate with the lowest plurality score.
    
    .. important::
        If there is  more than one candidate with the fewest number of first-place votes, then *all* such candidates are removed from the profile. 
    

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        Related functions:  :func:`pref_voting.iterative_methods.benham_put`

    """

    # need the total number of all candidates in a profile to check when all candidates have been removed   
    num_cands = profile.num_cands 
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    cw = profile.condorcet_winner(curr_cands = [c for c in profile.candidates if not isin(cands_to_ignore, c)])
    
    winners = [cw] if cw is not None else list()
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    while len(winners) == 0:
        plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates 
                            if  not isin(cands_to_ignore,c)}  
        min_plurality_score = min(plurality_scores.values())
        lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                             if  plurality_scores[c] == min_plurality_score])

        # remove cands with lowest plurality score
        cands_to_ignore = np.concatenate((cands_to_ignore, lowest_first_place_votes), axis=None)
        if len(cands_to_ignore) == num_cands: # removed all of the candidates 
            winners = sorted(lowest_first_place_votes)
        else:
            cw = profile.condorcet_winner([c for c in profile.candidates if not isin(cands_to_ignore, c)])
            if cw is not None: 
                winners = [cw]

    return sorted(winners)

@vm(name = "Benham TB",
    input_types=[ElectionTypes.PROFILE])
def benham_tb(profile, curr_cands = None, tie_breaker = None):
    """Benham (``benham``) with tie breaking:  If there is  more than one candidate with the fewest number of first-place votes, then remove the candidate with lowest in the tie_breaker ranking from the profile.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        tie_breaker (List[int]): A list of the candidates in the profile to be used as a tiebreaker.

    Returns: 
        A sorted list of candidates
   
    """

    # the tie_breaker is any linear order (i.e., list) of the candidates
    tb = tie_breaker if tie_breaker is not None else list(range(profile.num_cands))
    
    # need the total number of all candidates in a profile to check when all candidates have been removed   
    num_cands = profile.num_cands 
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    cw = profile.condorcet_winner(curr_cands = [c for c in profile.candidates if not isin(cands_to_ignore, c)])

    winners = [cw] if cw is not None else list()
    
    while len(winners) == 0:
        plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) for c in candidates if not isin(cands_to_ignore,c)}  
        min_plurality_score = min(plurality_scores.values())
        lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                             if plurality_scores[c] == min_plurality_score])
        
        cand_to_remove = lowest_first_place_votes[0]
        for c in lowest_first_place_votes[1:]: 
            if tb.index(c) < tb.index(cand_to_remove):
                cand_to_remove = c

        # remove cands with lowest plurality winners
        cands_to_ignore = np.concatenate((cands_to_ignore, cand_to_remove), axis=None)
        if len(cands_to_ignore) == num_cands: #all the candidates where removed
            winners = sorted(lowest_first_place_votes)
        else:
            cw = profile.condorcet_winner(curr_cands = [c for c in profile.candidates if not isin(cands_to_ignore, c)])
            if cw is not None: 
                winners = [cw]
    return sorted(winners)


@vm(name = "Benham PUT",
    input_types=[ElectionTypes.PROFILE])
def benham_put(profile, curr_cands = None):
    """Benham (:func:`benham`) with parallel universe tie-breaking (PUT), defined recursively: if there is a Condorcet winner, that candidate is the Benham-PUT winner; otherwise a candidate x is a Benham-PUT winner if there is some candidate y with minimal plurality score such that after removing y from the profile, x is a Benham-PUT winner.
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. warning:: 
        This will take a long time on profiles with many candidates having the same plurality scores.

    """
    candidates = profile.candidates if curr_cands is None else curr_cands 

    cw = profile.condorcet_winner(candidates)
    if cw is not None:
        return [cw]
    
    plurality_scores = profile.plurality_scores(candidates)
    worst_score = min(plurality_scores.values())

    cands_to_remove = [cand for cand, value in plurality_scores.items() if value == worst_score]
    
    winners = []
    for cand_to_remove in cands_to_remove:
        new_winners = benham_put(profile, curr_cands = [c for c in candidates if not c == cand_to_remove])
        winners = winners + new_winners
    
    return sorted(set(winners))

def iterated(vm):
    """Iteratively restrict the set of candidates to the vm winners until reaching a fixpoint.

    Args:
        vm (VotingMethod): A voting method.

    Returns:
        A voting method that iterates vm.

    """

    def _vm(edata, curr_cands = None):

        candidates = edata.candidates if curr_cands is None else curr_cands

        vm_ws = vm(edata, curr_cands=candidates)

        while not vm_ws == candidates:
            candidates = vm_ws
            vm_ws = vm(edata, curr_cands=candidates)

        return vm_ws

    return VotingMethod(_vm, name=f"Iterated {vm.name}")

def tideman_alternative(vm):
    """Given a voting method vm, returns a voting method that restricts the profile to the set of vm winners, then eliminates all the candidate with the fewest first-place votes, and then repeats until there is only one vm winner. If at some stage all remaining candidates are tied for the fewest number of first-place votes, then all remaining candidates win.

    Args:
        vm (VotingMethod): A voting method.

    Returns:
        The Tideman Alternative PUT version of vm.

    """
    
    def _ta(profile, curr_cands = None):

        candidates = profile.candidates if curr_cands is None else curr_cands 
    
        vm_ws = vm(profile, curr_cands = candidates)

        plurality_scores = profile.plurality_scores(vm_ws)
        worst_score = min(plurality_scores.values())

        cands_to_remove = [cand for cand, value in plurality_scores.items() if value == worst_score]
        
        if len(cands_to_remove) == len(vm_ws):
            return vm_ws
            
        else:
            return _ta(profile, curr_cands = [c for c in candidates if not c in cands_to_remove])
        
    _ta.__name__ = f"tideman_alternative_{vm.__name__}"
    return VotingMethod(_ta, name=f"Tideman Alternative {vm.name}")

tideman_alternative_smith = tideman_alternative(top_cycle)
tideman_alternative_smith.load_properties()
tideman_alternative_smith.input_types = [ElectionTypes.PROFILE]

tideman_alternative_gocha = tideman_alternative(gocha)
tideman_alternative_gocha.load_properties()
tideman_alternative_gocha.input_types = [ElectionTypes.PROFILE]

def tideman_alternative_put(vm):
    """Given a voting method vm, returns a voting method that restricts the profile to the set of vm winners, then eliminates the candidate with the fewest first-place votes, and then repeats until there is only one vm winner. Parallel-universe tiebreaking is used when there are multiple candidates with the fewest first-place votes.

    Args:
        vm (VotingMethod): A voting method.

    Returns:
        The Tideman Alternative PUT version of vm.

    """
    
    def _ta(profile, curr_cands = None):

        candidates = profile.candidates if curr_cands is None else curr_cands 
    
        vm_ws = vm(profile, curr_cands = candidates)

        if len(vm_ws) == 1:
            return vm_ws
    
        else: 
            plurality_scores = profile.plurality_scores(vm_ws)
            worst_score = min(plurality_scores.values())
            cands_to_remove = [cand for cand, value in plurality_scores.items() if value == worst_score]
        
            winners = []
            for cand_to_remove in cands_to_remove:
                additional_winners = _ta(profile, curr_cands = [c for c in candidates if not c == cand_to_remove])
                winners = winners + additional_winners

        return sorted(set(winners))
    
    _ta.__name__ = f"tideman_alternative_{vm.__name__}_put"
    return VotingMethod(_ta, name=f"Tideman Alternative {vm.name} PUT")


tideman_alternative_smith_put = tideman_alternative_put(top_cycle)
tideman_alternative_smith_put.load_properties()
tideman_alternative_smith_put.input_types = [ElectionTypes.PROFILE]

tideman_alternative_gocha_put = tideman_alternative_put(gocha)
tideman_alternative_gocha_put.load_properties()
tideman_alternative_gocha_put.input_types = [ElectionTypes.PROFILE]


@vm(name = "Woodall",
    input_types=[ElectionTypes.PROFILE])
def woodall(profile, curr_cands = None):
    """
    If there is a single member of the Smith Set (i.e., a Condorcet winner) then that candidate is the winner.  If there the Smith Set contains more than one candidate, then remove all candidates that are ranked first by the fewest number of voters.  Continue removing candidates with the fewest number first-place votes until there is a single member of the originally Smith Set remaining.  
    
    .. important::
        If there is  more than one candidate with the fewest number of first-place votes, then *all* such candidates are removed from the profile. 
    

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        Related functions:  :func:`pref_voting.iterative_methods.instant_runoff`

           
    """

    # need the total number of all candidates in a profile to check when all candidates have been removed   
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.empty(0) if curr_cands is None else np.array([c for c in profile.candidates if c not in curr_cands])

    s_set = top_cycle(profile, curr_cands=candidates)

    if len(s_set) == 1:
        return s_set
    
    rs, rcounts = profile.rankings_counts # get all the ranking data
    
    winners = []

    while len(winners) == 0:
        plurality_scores = {c: _num_rank_first(rs, rcounts, cands_to_ignore, c) 
                            for c in candidates if  not isin(cands_to_ignore,c)}  
        min_plurality_score = min(plurality_scores.values())
        lowest_first_place_votes = np.array([c for c in plurality_scores.keys() 
                                             if  plurality_scores[c] == min_plurality_score])

        remaining_cands_in_smith_set = [c for c in candidates if not isin(cands_to_ignore,c) and isin(np.array(s_set), c)]

        # remove cands with lowest plurality score
        new_cands_to_ignore = np.concatenate((cands_to_ignore, lowest_first_place_votes), axis=None)
        
        new_remaining_cands_in_smith_set = [c for c in candidates if not isin(new_cands_to_ignore,c) and isin(np.array(s_set), c)]

        if len(new_remaining_cands_in_smith_set) == 0: 
            winners = remaining_cands_in_smith_set
        
        if len(new_remaining_cands_in_smith_set) == 1:
            winners = new_remaining_cands_in_smith_set 
     
        cands_to_ignore = new_cands_to_ignore

    return sorted(winners)

@vm(name = "Knockout Voting",
    input_types=[ElectionTypes.PROFILE])
def knockout(profile, curr_cands=None):
    """Find the two candidates in curr_cands with the lowest and second lowest Borda scores among any candidates in curr_cands. Then remove from curr_cands whichever one loses to the other in a head-to-head majority comparison. Repeat this process, always using the original Borda score (i.e., the Borda scores calculated with respect to all candidates in the profile, not with respect to curr_cands as for Baldwin and Nanson) until only one candidate remains in curr_cands. Parallel universe tie-breaking (PUT) is used when there are ties in lowest or second lowest Borda scores.

    .. note::
        Proposed by Edward B. Foley (with unspecified handling of ties in Borda scores, so PUT is used here as an example).
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    """
    candidates = profile.candidates if curr_cands is None else curr_cands 
    
    if len(candidates) == 1:
        return candidates

    # Key point: use global Borda score, calculated with respect to the full profile, not just the candidates in curr_cands
    borda_scores = profile.borda_scores()
    min_borda_score = min([borda_scores[c] for c in candidates])
    cands_with_lowest_borda_score = [c for c in candidates if c in borda_scores.keys() and borda_scores[c] == min_borda_score]
    
    winners = []

    # If multiple candidates tie for lowest Borda score, consider all head-to-head matchups of these candidates
    if len(cands_with_lowest_borda_score) > 1:
        for c1 in cands_with_lowest_borda_score:
            for c2 in cands_with_lowest_borda_score:
                if c1 != c2:
                    # If c1 has a non-negative margin over c2, then remove c2 from curr_cands and calculate the winning set
                    # Take the union over all such winning sets as the ultimate winning set
                    if profile.margin(c1, c2) >= 0:
                        new_winners = knockout(profile, curr_cands = [c for c in candidates if not c == c2])
                        winners = winners + new_winners
    
    # If there is a candidate with the uniquely lowest Borda score
    if len(cands_with_lowest_borda_score) == 1:
        cand_with_lowest_borda_score = cands_with_lowest_borda_score[0]

        # There may be multiple candidates with the second lowest Borda score
        second_lowest_borda_score = min([borda_scores[c] for c in candidates if c not in cands_with_lowest_borda_score])
        cands_with_second_lowest_borda_score = [c for c in candidates if c in borda_scores.keys() and borda_scores[c] == second_lowest_borda_score]

        # Consider all head-to-head matchups between the candidate with the lowest Borda score and the candidates with the second lowest Borda score
        for c2 in cands_with_second_lowest_borda_score:

            # If a candidate with second lowest Borda score has a non-negative margin over the candidate with the lowest Borda score, 
            # then remove the latter from curr_cands and calculate the winning set
            if profile.margin(c2, cand_with_lowest_borda_score) >= 0:
                new_winners = knockout(profile, curr_cands = [c for c in candidates if not c == cand_with_lowest_borda_score])
                winners = winners + new_winners

            # If the candidate with the lowest Borda score has a positive margin over a candidate with the second lowest Borda score, 
            # then remove the latter from curr_cands and calculate the winning set
            if profile.margin(cand_with_lowest_borda_score, c2) > 0:
                new_winners = knockout(profile, curr_cands = [c for c in candidates if not c == c2])
                winners = winners + new_winners

    return sorted(set(winners))

@vm(name="Plurality Veto",
    input_types=[ElectionTypes.PROFILE])
def plurality_veto(profile, curr_cands=None, voter_order=None):
    """Returns the winner using the Plurality Veto method of Kizilkaya and Kempe (https://arxiv.org/abs/2305.19632). 

    The method works as follows:
    1. Assign initial scores to candidates equal to their plurality scores
    2. Process voters one by one in the given order
    3. Each voter decrements the score of their bottom choice among non-eliminated candidates
    4. A candidate is eliminated when their score reaches zero
    5. The winner is the last remaining candidate

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in curr_cands
        voter_order (List[int], optional): List of voters in the order to process them. If None, uses range(len(profile.rankings))

    Returns:
        A sorted list of candidates

    warning::
        If no voter order is specified, the method uses the default order of voter rankings in the profile. Note that anonymizing a profile changes the order of voter rankings.
    """
    candidates = profile.candidates if curr_cands is None else curr_cands

    # Initialize scores as plurality scores
    scores = profile.plurality_scores(curr_cands=candidates)

    # If no voter order specified, use default order
    if voter_order is None:
        voter_order = list(range(profile.num_voters))

    # Track non-eliminated candidates and last remaining
    active_candidates = set(candidates)
    last_remaining = None  # Track the last remaining candidate

    # Process each voter
    for voter in voter_order:
        # Get remaining candidates with positive scores
        remaining = {c for c in active_candidates if scores[c] > 0}
        if not remaining:
            # If all remaining candidates have 0 scores, return the last remaining
            return [last_remaining] if last_remaining is not None else sorted(active_candidates)

        # If only one candidate remains with positive score, they are the winner
        if len(remaining) == 1:
            return sorted(remaining)

        # Get voter's bottom choice among remaining candidates
        ranking = profile.rankings[voter]
        # Find the last ranked candidate among remaining ones
        bottom = next(c for c in reversed(ranking) if c in remaining)

        # Decrement score
        scores[bottom] -= 1
        if scores[bottom] == 0:
            active_candidates.remove(bottom)
            last_remaining = bottom

    # Return the last remaining candidate if there was one,
    # otherwise return candidates with highest remaining score
    if last_remaining is not None:
        return [last_remaining]
    else:
        max_score = max(scores.values())
        return sorted([c for c in candidates if scores[c] == max_score])

def plurality_veto_with_explanation(profile, curr_cands=None, voter_order=None):
    """Returns the winner using the Plurality Veto method, with a detailed explanation of the process.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to curr_cands
        voter_order (List[int], optional): List of voters in the order to process them. If None, uses range(len(profile.rankings))

    Returns:
        tuple: A tuple containing (winner list, explanation string)

    warning::
        If no voter order is specified, the method uses the default order of voter rankings in the profile. Note that anonymizing a profile changes the order of voter rankings.
    """
    curr_cands = profile.candidates if curr_cands is None else curr_cands
    scores = profile.plurality_scores(curr_cands=curr_cands)

    if voter_order is None:
        voter_order = list(range(profile.num_voters))

    explanation = [
        "Initial plurality scores: " + str(dict(scores)),
    ]

    # Note any candidates eliminated due to zero initial plurality scores
    zero_initial = [c for c in curr_cands if scores[c] == 0]
    if zero_initial:
        explanation.append(f"Candidates eliminated due to zero initial plurality score: {sorted(zero_initial)}")
    explanation.append("")  # Add blank line

    active_candidates = set(curr_cands)
    last_remaining = None

    # Add initially eliminated candidates
    for c in zero_initial:
        active_candidates.remove(c)
        last_remaining = c

    for step, voter in enumerate(voter_order):
        remaining = {c for c in active_candidates if scores[c] > 0}
        if not remaining:
            explanation.append("All remaining candidates have score 0")
            if last_remaining is not None:
                explanation.append(f"Winners are candidates [{last_remaining}] (highest remaining scores)")
                return [last_remaining], "\\n".join(explanation)
            else:
                winners = sorted(active_candidates)
                explanation.append(f"Winners are candidates {winners} (highest remaining scores)")
                return winners, "\\n".join(explanation)

        # If only one candidate remains with positive score, they are the winner
        if len(remaining) == 1:
            winners = sorted(remaining)
            explanation.append(f"Only one candidate remains with positive score")
            explanation.append(f"Winners: {winners} (highest remaining scores)")
            return winners, "\\n".join(explanation)

        ranking = profile.rankings[voter]
        # Filter ranking to show only active candidates
        active_ranking = [c for c in ranking if c in remaining]
        bottom = next(c for c in reversed(ranking) if c in remaining)

        explanation.append(f"Step {step + 1}:")
        explanation.append(f"Voter {voter} (active candidates in ranking: {active_ranking}) vetoes {bottom}")

        scores[bottom] -= 1
        explanation.append(f"Scores after veto: {dict({c: s for c, s in scores.items() if c in remaining})}")

        if scores[bottom] == 0:
            active_candidates.remove(bottom)
            last_remaining = bottom
            explanation.append(f"Candidate {bottom} eliminated!")
        explanation.append("")

    if last_remaining is not None:
        explanation.append(f"Winners: [{last_remaining}] (highest remaining scores)")
        return [last_remaining], "\\n".join(explanation)
    else:
        max_score = max(scores.values())
        winners = sorted([c for c in curr_cands if scores[c] == max_score])
        explanation.append(f"Winners: {winners} (highest remaining scores)")
        return winners, "\\n".join(explanation)
    
@vm(name="Consensus Builder",
    input_types=[ElectionTypes.PROFILE])
def consensus_builder(profile, curr_cands=None, consensus_building_ranking=None, beta=0.5):

    """Deterministic version of the Random Consensus Builder due to Charikar et al. (https://arxiv.org/abs/2306.17838).

    The method processes candidates in reverse order of the consensus building ranking. When processing
    candidate i, it eliminates any candidate j above i in the consensus building ranking if a large enough fraction of voters (>= beta) prefer i to j. The winner is the last candidate that gets processed.

    Args:
        profile (Profile): An anonymous profile of linear orders
        curr_cands (List[int], optional): Candidates to consider. Defaults to all candidates if not provided.
        consensus_building_ranking (List[int]): The ranking to use as the consensus builder. If not provided, uses the lexicographically first ranking of curr_cands.
        beta (float): Threshold for elimination (default 0.5). When processing candidate i, eliminates a candidate j above i in the consensus building ranking if the proportion of voters preferring i to j is >= beta

    Returns:
        list: List containing the winning candidate

    .. seealso::
        :meth:`pref_voting.probabilistic_methods.random_consensus_builder`
        :meth:`pref_voting.stochastic_methods.random_consensus_builder_st`
    """

    if curr_cands is None:
        curr_cands = profile.candidates

    if consensus_building_ranking is None:
        consensus_building_ranking = sorted(curr_cands)

    # all candidates in curr_cands must be in consensus_building_ranking
    assert len([c for c in curr_cands if c not in consensus_building_ranking]) == 0

    eliminated = set()
    last_processed = None

    for i in reversed(consensus_building_ranking):

        if i not in curr_cands or i in eliminated:
            continue

        for j in consensus_building_ranking:
            if j == i or j not in curr_cands or j in eliminated:
                continue

            if consensus_building_ranking.index(j) < consensus_building_ranking.index(i):
                support_ratio = profile.support(i, j) / profile.num_voters
                if support_ratio >= beta:
                    eliminated.add(j)

        last_processed = i

    return [last_processed]

iterated_vms_with_explanation = [
    instant_runoff_with_explanation,
    coombs_with_explanation,
    plurality_with_runoff_put_with_explanation,
    baldwin_with_explanation,
    strict_nanson_with_explanation,
    weak_nanson_with_explanation,
    iterated_removal_cl_with_explanation,
    plurality_veto_with_explanation
]


'''
    File: prob_voting_method.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: April 14, 2024
    
    The ProbabilisticVotingMethod class and helper functions for probabilistic voting methods.
'''

import functools
import numpy as np
import inspect

class ProbVotingMethod(object): 
    """
    A class to add functionality to probabilistic voting methods 

    Args:
        pvm (function): An implementation of a probabilistic voting method. The function should accept any type of profile, and a keyword parameter ``curr_cands`` to find the winner after restricting to ``curr_cands``. 
        name (string): The human-readable name of the social welfare function.

    Returns:
        A dictionary that represents the probability on the set of candidates.

    """
    def __init__(self, pvm, name = None): 
        
        self.pvm = pvm
        self.name = name
        self.algorithm = None

        functools.update_wrapper(self, pvm)   

    def __call__(self, edata, curr_cands = None, **kwargs):

        if (curr_cands is not None and len(curr_cands) == 0) or len(edata.candidates) == 0: 
            return {}
        return self.pvm(edata, curr_cands = curr_cands, **kwargs)
        
    def support(self, edata, curr_cands = None, **kwargs): 
        """
        Return the sorted list of the set of candidates that have non-zero probability. 
        """

        if (curr_cands is not None and len(curr_cands) == 0) or len(edata.candidates) == 0: 
            return []
        prob =  self.pvm(edata, curr_cands = curr_cands, **kwargs)
        return sorted([c for c,pr in prob.items() if pr > 0])
    
    def choose(self, edata, curr_cands = None, **kwargs): 
        """
        Return a randomly chosen element according to the probability. 
        """

        prob =  self.pvm(edata, curr_cands = curr_cands, **kwargs)
        
        # choose a candidate according to the probability distribution prob
        cands = list(prob.keys())
        probs = [prob[c] for c in cands]

        return np.random.choice(cands, p = probs)
    
    def display(self, edata, curr_cands = None, cmap = None, **kwargs): 
        """
        Display the winning set of candidates.
        """
 
        cmap = cmap if cmap is not None else edata.cmap

        prob = self.__call__(edata, curr_cands = curr_cands, **kwargs)

        if prob is None:  # some voting methods may return None if, for instance, it is taking long to compute the winner.
            print(f"{self.name} probability is not available")
        else: 
            w_str = f"{self.name} probability is " 
            print(w_str + "{" + ", ".join([f"{str(cmap[c])}: {round(pr,3)}" for c,pr in prob.items()]) + "}")
        

    def set_name(self, new_name):
        """Set the name of the social welfare function."""

        self.name = new_name

    def set_algorithm(self, algorithm):
        """
        Set the algorithm for the voting method if 'algorithm' is an accepted keyword parameter.

        Args:
            algorithm: The algorithm to set for the voting method.
        """
        params = inspect.signature(self.pvm).parameters
        if 'algorithm' in params and params['algorithm'].kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]:
            self.algorithm = algorithm
        else:
            raise ValueError(f"The method {self.name} does not accept 'algorithm' as a parameter.")

    def __str__(self): 
        return f"{self.name}"

def pvm(name = None):
    """
    A decorator used when creating a social welfare function. 
    """
    def wrapper(f):
        return ProbVotingMethod(f, name=name)
    return wrapper



"""
    File: gen_profiles.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: December 7, 2020
    Updated: May 25, 2025
    
    Functions to generate profiles

"""

from itertools import combinations
from pref_voting.profiles import Profile
from pref_voting.generate_spatial_profiles import generate_spatial_profile
from pref_voting.generate_utility_profiles import linear_utility
import numpy as np 
import math
import random
from scipy.stats import gamma
from itertools import permutations
from pref_voting.helper import weak_compositions, weak_orders

from pref_voting.profiles_with_ties import ProfileWithTies
from ortools.linear_solver import pywraplp
from prefsampling.ordinal import impartial, impartial_anonymous, urn, plackett_luce, didi, stratification, single_peaked_conitzer, single_peaked_walsh, single_peaked_circle, single_crossing, euclidean, mallows

from prefsampling.core.euclidean import EuclideanSpace
from collections import Counter

# ############
# wrapper functions to interface with preflib tools for generating profiles
# ############


# Given the number m of candidates and a phi in [0,1], 
# compute the expected number of swaps in a vote sampled 
# from the Mallows model
def find_expected_number_of_swaps(num_candidates, phi):
    res = phi * num_candidates / (1 - phi)
    for j in range(1, num_candidates + 1):
        res = res + (j * (phi**j)) / ((phi**j) - 1)
    return res


# Given the number m of candidates and a absolute number of 
# expected swaps exp_abs, this function returns a value of 
# phi such that in a vote sampled from Mallows model with 
# this parameter the expected number of swaps is exp_abs
def phi_from_relphi(num_candidates, relphi=None, seed=None):

    rng = np.random.default_rng(seed)
    if relphi is None:
        relphi = rng.uniform(0.001, 0.999)
    if relphi == 1:
        return 1
    exp_abs = relphi * (num_candidates * (num_candidates - 1)) / 4
    low = 0
    high = 1
    while low <= high:
        mid = (high + low) / 2
        cur = find_expected_number_of_swaps(num_candidates, mid)
        if abs(cur - exp_abs) < 1e-5:
            return mid
        # If x is greater, ignore left half
        if cur < exp_abs:
            low = mid

        # If x is smaller, ignore right half
        elif cur > exp_abs:
            high = mid

    # If we reach here, then the element was not present
    return -1

# Return a list of phis from the relphi value
def phis_from_relphi(num_candidates, num, relphi=None, seed=None):

    rng = np.random.default_rng(seed)
    if relphi is None:
        relphis = rng.uniform(0.001, 0.999, size=num)
    else: 
        relphis = [relphi] * num
    
    return [phi_from_relphi(num_candidates, relphi=relphis[n]) for n in range(num)]


def get_rankings(num_candidates, num_voters, **kwargs): 
    """
    Get the rankings for a given number of candidates and voters using
    the [prefsampling library](https://comsoc-community.github.io/prefsampling/index.html). 

    Args:
        num_candidates (int): The number of candidates.
        num_voters (int): The number of voters.
        kwargs (dict): Any parameters for the probability model.
    
    Returns:
        list: A list of rankings.
    """

    if 'probmodel' in kwargs:
        probmodel = kwargs['probmodel']
    elif 'probmod' in kwargs: # for backwards compatibility
        probmodel = kwargs['probmod']
    else: 
        probmodel = "impartial"

    if 'seed' in kwargs:
        seed = kwargs['seed']
    else: 
        seed = None

    if probmodel == "IC" or probmodel == 'impartial': 
        
        rankings = impartial(num_voters, 
                             num_candidates, 
                             seed=seed) 
    
    elif probmodel == "IAC" or probmodel == 'impartial_anonymous': 
        
        rankings = impartial_anonymous(num_voters, 
                                       num_candidates, 
                                       seed=seed)
    elif probmodel == "MALLOWS" or probmodel == 'mallows':

        impartial_central_vote = True
        if 'phi' in kwargs: 
            phi = kwargs['phi']
        else:
            phi = 1.0
            
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)


    elif probmodel == "MALLOWS-0.8":

        phi = 0.8
        impartial_central_vote = True           
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)
        
    elif probmodel == "MALLOWS-0.2":

        phi = 0.2
        impartial_central_vote = True           
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)

    elif probmodel == "MALLOWS-R":

        rng = np.random.default_rng(seed)
        phi = rng.uniform(0.001, 0.999)
        impartial_central_vote = True
            
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)

    elif probmodel == "MALLOWS-RELPHI":

        impartial_central_vote = True
        if 'relphi' in kwargs: 
            relphi = kwargs['relphi']
        else:
            relphi = None
            
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        phi = phi_from_relphi(num_candidates, relphi=relphi, seed=seed)

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)

    elif probmodel == "MALLOWS-RELPHI-0.375":
        
        relphi = 0.375
        impartial_central_vote = True
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        phi = phi_from_relphi(num_candidates, relphi=relphi, seed=seed)

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)


    elif probmodel == "MALLOWS-RELPHI-R":
            
        impartial_central_vote = True
        if 'normalise_phi' in kwargs: 
            normalise_phi = kwargs['normalise_phi']
        else:
            normalise_phi = False

        if 'central_vote' in kwargs: 
            central_vote = kwargs['central_vote']
            impartial_central_vote = False
        else:
            central_vote = None

        phi = phi_from_relphi(num_candidates, relphi=None, seed=seed)

        rankings = mallows(num_voters,
                           num_candidates, 
                           phi,
                           normalise_phi=normalise_phi,
                           central_vote=central_vote,
                           impartial_central_vote=impartial_central_vote,
                           seed=seed)


    elif probmodel == "URN" or probmodel == 'urn': 

        if 'alpha' in kwargs: 
            alpha = kwargs['alpha']
        else:
            alpha = 0.0
            
        rankings = urn(num_voters,
                       num_candidates, 
                       alpha,
                       seed=seed)

    elif probmodel == "URN-10":
        
        alpha = 10
        rankings = urn(num_voters,
                       num_candidates, 
                       alpha,
                       seed=seed)
    
    elif probmodel == "URN-0.3":
        
        alpha = round(math.factorial(num_candidates) * 0.3)
        rankings = urn(num_voters,
                       num_candidates, 
                       alpha,
                       seed=seed)
        
    elif probmodel == "URN-R":
        
        rng = np.random.default_rng(seed)
        alpha = round(math.factorial(num_candidates) * gamma.rvs(0.8, random_state=rng))
        rankings = urn(num_voters,
                       num_candidates,
                       alpha,
                       seed=seed)
        
    elif probmodel == "plackett_luce":
        
        if 'alphas' not in kwargs:
            raise ValueError("Error: alphas parameter missing.  A value must be specified for each candidate indicating their relative quality.")
            #RaiseValueError()
        else:
            alphas = kwargs['alphas']

        rankings = plackett_luce(num_voters,
                                       num_candidates, 
                                       alphas,
                                       seed=seed)
        
    elif probmodel == "didi":
        
        if 'alphas' not in kwargs:
            raise ValueError("Error: alphas parameter missing.  A value must be specified for each candidate indicating each candidate's quality.")
        else:
            alphas = kwargs['alphas']

        rankings = didi(num_voters,
                        num_candidates, 
                        alphas,
                        seed=seed)
        
    elif probmodel == "stratification":
        
        if 'weight' not in kwargs:
            raise ValueError("Error: weight parameter missing.  The weight parameter specifies the size of the upper class of candidates.")
        else:
            weight = kwargs['weight']

        rankings = stratification(num_voters,
                                  num_candidates, 
                                  weight,
                                  seed=seed) 
    
    elif probmodel == "single_peaked_conitzer":
        
        rankings = single_peaked_conitzer(num_voters,
                                          num_candidates, 
                                          seed=seed) 
    
    elif probmodel == "SinglePeaked" or probmodel == "single_peaked_walsh":
        
        rankings = single_peaked_walsh(num_voters,
                                       num_candidates, 
                                       seed=seed) 

    elif probmodel == "single_peaked_circle":
        
        rankings = single_peaked_circle(num_voters,
                                        num_candidates, 
                                        seed=seed)       

    elif probmodel == "single_crossing":
        
        rankings = single_crossing(num_voters,
                                   num_candidates, 
                                   seed=seed) 
        
    elif probmodel == "euclidean":
        
        euclidean_spaces = {
            "gaussian_ball": EuclideanSpace.GAUSSIAN_BALL,
            "gaussian_cube": EuclideanSpace.GAUSSIAN_CUBE,
            "unbounded_gaussian": EuclideanSpace.UNBOUNDED_GAUSSIAN,
            "uniform_ball": EuclideanSpace.UNIFORM_BALL,
            "uniform_cube": EuclideanSpace.UNIFORM_CUBE,
            "uniform_sphere": EuclideanSpace.UNIFORM_SPHERE,
        }

        if 'space' in kwargs:
            space = kwargs['space']
        else:
            space = "uniform_ball"

        if 'dimension' in kwargs:
            dimension = kwargs['dimension']
        else:
            dimension = 2

        rankings = euclidean(num_voters,
                             num_candidates, 
                             voters_positions=euclidean_spaces[space],
                             candidates_positions=euclidean_spaces[space],
                             num_dimensions=dimension, 
                             seed=seed) 
    
    else: 
        raise ValueError("Error: The probability model is not recognized.")
        
    return rankings

def generate_profile(num_candidates, 
                     num_voters, 
                     anonymize=False,
                     num_profiles=1,
                     **kwargs): 
    """
    Generate profiles using the prefsampling library.

    Args:
        num_candidates (int): The number of candidates.
        num_voters (int): The number of voters.
        anonymize (bool): If True, anonymize the profiles.
        num_profiles (int): The number of profiles to generate.
        kwargs (dict): Any parameters for the probability model.

    Returns:
        list: A list of profiles or a single profile if num_profiles is 1.  
    """
            
    profs = [Profile(get_rankings(num_candidates,
                                  num_voters, 
                                  **kwargs))  
                                  for _ in range(num_profiles)]
    
    if anonymize: 
        profs = [prof.anonymize() for prof in profs]
        
    return profs[0] if num_profiles == 1 else profs

def generate_profile_with_groups(
        num_candidates, 
        num_voters, 
        probmodels, 
        weights=None,
        seed=None, 
        num_profiles=1, 
        anonymize=False):
    
    """
    Generate profiles with groups of voters generated from different probability models.
    The probability of selecting a probability model is proportional its weight in the list weight.

    Args:
        num_candidates (int): The number of candidates.
        num_voters (int): The number of voters.
        probmodels (list): A list of dictionaries specifying a probability model.
        weights (list): A list of weights for each probability model.
        seed (int): The random seed.
        num_profiles (int): The number of profiles to generate.
        anonymize (bool): If True, anonymize the profiles.
    """
    if weights is None:
        weights = [1] * len(probmodels)
    
    assert len(weights)==len(probmodels), "The number of weights must be equal to the number of probmodels"

    probs = [w / sum(weights) for w in weights]
    
    rng = np.random.default_rng(seed)

    profs = list()
    for _ in range(num_profiles):
        selected_probmodels = rng.choice(probmodels, num_voters, p=probs)

        selected_probmodels_num = Counter([tuple((k,v) if type(v) != list else (k, tuple(v)) for k,v in pm.items()) for pm in selected_probmodels])

        rankings = list()
        for pm_data, nv in selected_probmodels_num.items():
            rankings = rankings + list(get_rankings(num_candidates, nv, **dict(pm_data)))

        prof = Profile(rankings)
        if anonymize: 
            prof = prof.anonymize()
        profs.append(prof)

    return profs[0] if num_profiles == 1 else profs

####
# Enumerating profiles
####

def enumerate_anon_profile(num_cands, num_voters):
    """A generator that enumerates all anonymous profiles with num_cands candidates and num_voters voters.

    Args:
        num_cands (int): Number of candidates.
        num_voters (int): Number of voters.

    Yields:
        Profile: An anonymous profile.
    """
    
    ballot_types = list(permutations(range(num_cands)))
    num_ballot_types = len(ballot_types)

    for comp in weak_compositions(num_voters, num_ballot_types):
        instantiated_ballot_types = [ballot_types[idx] for idx, i in enumerate(comp) if i != 0]
        nonzerocomp = [i for i in comp if i != 0]
        yield Profile(instantiated_ballot_types, rcounts = nonzerocomp)

def canonical_ballot_multiset(profile: Profile) -> tuple:
    """
    Lexicographically minimal multiset of (ranking, count) pairs across
    all candidate permutations.
    """
    m = profile.num_cands
    rankings, counts = profile.rankings_counts
    counts = counts.astype(int)
    
    best = None
    
    # For each ranking that could potentially be the "canonical" one (0,1,2,...)
    for r in rankings:
        # Find the permutation that maps r to (0,1,2,...,m-1)
        perm = {int(r[i]): i for i in range(m)}
        
        # Apply this inverse permutation to all rankings
        canon = tuple(sorted(
            (tuple(perm[c] for c in ranking.tolist()), int(k))
            for ranking, k in zip(rankings, counts)
        ))
        
        if best is None or canon < best:
            best = canon
    
    return best

def enumerate_anon_neutral_profile(num_cands: int, num_voters: int):
    """
    A generator that yields one representative per neutrality-orbit of anonymous profiles.
    """
    seen = set()
    for prof in enumerate_anon_profile(num_cands, num_voters):
        key = canonical_ballot_multiset(prof)
        if key not in seen:
            seen.add(key)
            yield prof


def enumerate_anon_profile_with_ties(num_cands, num_voters):
    """A generator that enumerates all anonymous profiles--allowing ties in ballots--with num_cands candidates and num_voters voters

    Args:
        num_cands (int): Number of candidates.
        num_voters (int): Number of voters.

    Yields:
        ProfileWithTies: An anonymous profile.
    """
    
    ballot_types = list(weak_orders(range(num_cands)))
    num_ballot_types = len(ballot_types)

    for comp in weak_compositions(num_voters, num_ballot_types):
        instantiated_ballot_types = [ballot_types[idx] for idx, i in enumerate(comp) if i != 0]
        nonzerocomp = [i for i in comp if i != 0]
        yield ProfileWithTies(instantiated_ballot_types, rcounts = nonzerocomp)

def _weakorder_to_levels(order):
    """Convert a weak order dict -> tuple of rank-levels."""
    if not order:
        return tuple()
    max_rank = max(order.values())
    return tuple(
        tuple(sorted(c for c, r in order.items() if r == lev))
        for lev in range(max_rank + 1)
    )

def _canonical_multiset_with_ties(profile):
    """Canonical key for a *ProfileWithTies* under candidate permutations."""
    m = profile.num_cands
    ballots, counts = profile.rankings_counts
    ballots = list(ballots)
    counts  = [int(k) for k in counts]

    best = None
    for perm in permutations(range(m)):
        relabel = dict(zip(range(m), perm))

        canon = tuple(sorted(
            (
                _weakorder_to_levels(
                    {relabel[c]: r for c, r in (
                        ballot if isinstance(ballot, dict) else ballot.rmap
                    ).items()}
                ),
                k,
            )
            for ballot, k in zip(ballots, counts)
        ))
        if best is None or canon < best:
            best = canon
    return best

def enumerate_anon_neutral_profile_with_ties(num_cands, num_voters):
    """A generator that yields one representative per neutrality-orbit of anonymous profiles allowing ties.
    """
    seen = set()
    for prof in enumerate_anon_profile_with_ties(num_cands, num_voters):
        key = _canonical_multiset_with_ties(prof)
        if key not in seen:
            seen.add(key)
            yield prof


####
# Generating ProfilesWithTies
####


def strict_weak_orders(A):
    if not A:  # i.e., A is empty
        yield []
        return
    for k in range(1, len(A) + 1):
        for B in combinations(A, k):  # i.e., all nonempty subsets B
            for order in strict_weak_orders(set(A) - set(B)):
                yield [B] + order


def generate_truncated_profile(
        num_cands, 
        num_voters, 
        max_num_ranked=3,
        probmod="IC"):
    
    """Generate a :class:`ProfileWithTies` with ``num_cands`` candidates and ``num_voters``.  
    The ballots will be truncated linear orders of the candidates.  Returns a :class:`ProfileWithTies` that uses extended strict preference (so all ranked candidates are strictly preferred to any candidate that is not ranked).

    Args:
        num_cands (int): The number of candidates to include in the profile. 
        num_voters (int): The number of voters to include in the profile.
        max_num_ranked (int, default=3): The maximum level to truncate the linear ranking. 
        probmod (str): optional (default "IC")

    Returns: 
        ProfileWithTies 

    :Example:

        .. exec_code::

            from pref_voting.generate_profiles import generate_truncated_profile

            prof = generate_truncated_profile(6, 7)
            prof.display()

            prof = generate_truncated_profile(6, 7, max_num_ranked=6)
            prof.display()

    :Possible Values of probmod:
    
    - "IC" (Impartial Culture): each randomly generated linear order of all candidates is truncated at a level from 1 to max_num_ranked, where the probability of truncating at level t is the number of truncated linear orders of length t divided by the number of truncated linear orders of length from 1 to max_num_ranked. Then a voter is equally likely to get any of the truncated linear orders of length from 1 to max_num_ranked.
    - "RT" (Random Truncation): each randomly generated linear order of all candidates is truncated at a level that is randomly chosen from 1 to max_num_ranked.
        
    """
    
    if max_num_ranked > num_cands:
        max_num_ranked = num_cands

    if probmod == "IC":
        num_rankings_of_length = dict()

        for n in range(1, max_num_ranked + 1):
            num_rankings_of_length[n] = 1
            for i in range(num_cands,num_cands-n, -1):
                num_rankings_of_length[n] *= i

        num_all_rankings = sum([num_rankings_of_length[n] for n in range(1, max_num_ranked + 1)])
        probabilities = [num_rankings_of_length[n] / num_all_rankings for n in range(1, max_num_ranked + 1)]

    lprof = generate_profile(num_cands, num_voters)
    
    rmaps = list()
    for r in lprof.rankings:

        if probmod == "RT":
            truncate_at = random.choice(range(1, max_num_ranked + 1))

        if probmod == "IC":
            truncate_at = random.choices(range(1, max_num_ranked + 1), weights=probabilities, k=1)[0]

        truncated_r = r[0:truncate_at]

        rmap = {c: _r + 1 for _r, c in enumerate(truncated_r)}

        rmaps.append(rmap)

    prof = ProfileWithTies(
        rmaps,
        cmap=lprof.cmap,
        candidates=lprof.candidates
    )
    prof.use_extended_strict_preference()
    return prof

####
# Generating Profile from ordinal margin graph
####

def minimal_profile_from_edge_order(cands, edge_order):
    """Given a list of candidates and a list of edges (positive margin edges only) in order of descending strength, find a minimal profile whose ordinal margin graph has that edge order.

    Args: 
        cands (list): list of candidates
        edge_order (list): list of edges in order of descending strength

    Returns:
        Profile: a profile whose ordinal margin graph has the given edge order
    """

    solver = pywraplp.Solver.CreateSolver("SAT")

    num_cands = len(cands)
    rankings = list(permutations(range(num_cands)))

    ranking_to_var = dict()
    infinity = solver.infinity()
    for ridx, r in enumerate(rankings): 
        _v = solver.IntVar(0.0, infinity, f"x{ridx}")
        ranking_to_var[r] = _v

    nv = solver.IntVar(0.0, infinity, "nv")
    equations = list()
    for c1 in cands: 
        for c2 in cands: 
            if c1 != c2: 
                if (c1,c2) in edge_order:
                    rankings_c1_over_c2 = [ranking_to_var[r] for r in rankings if r.index(c1) < r.index(c2)]
                    rankings_c2_over_c1 = [ranking_to_var[r] for r in rankings if r.index(c2) < r.index(c1)]
                    equations.append(sum(rankings_c1_over_c2) - sum(rankings_c2_over_c1) >= 1)

                for c3 in cands:
                    for c4 in cands:
                        if c3 != c4:
                            if (c1,c2) in edge_order and (c3,c4) in edge_order and edge_order.index((c1,c2)) < edge_order.index((c3,c4)):
                                rankings_c3_over_c4 = [ranking_to_var[r] for r in rankings if r.index(c3) < r.index(c4)]
                                rankings_c4_over_c3 = [ranking_to_var[r] for r in rankings if r.index(c4) < r.index(c3)]
                                equations.append(sum(rankings_c1_over_c2) - sum(rankings_c2_over_c1) >= sum(rankings_c3_over_c4) - sum(rankings_c4_over_c3) + 1)
                    
    equations.append(nv == sum(list(ranking_to_var.values())))

    for eq in equations: 
        solver.Add(eq)

    solver.Minimize(nv)

    status = solver.Solve()

    if status == pywraplp.Solver.INFEASIBLE:
        print("Error: Did not find a solution.")
        return None

    if status != pywraplp.Solver.OPTIMAL: 
        print("Warning: Did not find an optimal solution.")

    _ranks = list()
    _rcounts = list()

    for r,v in ranking_to_var.items(): 

        if v.solution_value() > 0: 
            _ranks.append(r)
            _rcounts.append(int(v.solution_value()))
            if not v.solution_value().is_integer(): 
                print("ERROR: Found non integer, ", v.solution_value())
                return None

    return Profile(_ranks, rcounts = _rcounts)


'''
    File: stochastic_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: November 22, 2024
    
    Implementations of voting methods that output winners stochastically (unlike probabilistic methods, which output a probability distribution in the form of a dictionary).
'''

from pref_voting.voting_method import *
from pref_voting.iterative_methods import consensus_builder
from pref_voting.probabilistic_methods import maximal_lottery, RaDiUS
import math

@vm(name="Random Consensus Builder (Stochastic)")
def random_consensus_builder_st(profile, curr_cands=None, beta=0.5):

    """Version of the Random Consensus Builder (RCB) voting method due to Charikar et al. (https://arxiv.org/abs/2306.17838) that actually chooses a winner stochastically rather than outputting a probability distribution.

    Args:

        profile (Profile): An anonymous profile of linear orders
        curr_cands (List[int], optional): Candidates to consider. Defaults to all candidates if not provided.
        beta (float): Threshold for elimination (default 0.5). When processing candidate i, eliminates a candidate j
                    above i in the consensus building ranking if the proportion of voters preferring i to j is >= beta

    Returns:
        A sorted list of candidates.

    .. seealso::
        :meth:`pref_voting.iterative_methods.consensus_builder`
        :meth:`pref_voting.probabilistic_methods.random_consensus_builder`

    """
    consensus_building_ranking = random.choice(profile.rankings)

    return consensus_builder(profile, curr_cands=curr_cands, consensus_building_ranking=consensus_building_ranking, beta=beta)

@vm(name="Maximal Lotteries mixed with Random Consensus Builder")
def MLRCB(profile, curr_cands=None, p = 1 / math.sqrt(2), B = math.sqrt(2) - 1/2):

    """With probability p, choose the winner from the Maximal Lotteries distribution. With probability 1-p, run the stochastic version of Random Consensus Builder with beta chosen uniformly from (1/2, B). Ths method comes from Theorem 4 of Charikar et al. (https://arxiv.org/abs/2306.17838).

    Args:

        profile (Profile): An anonymous profile of linear orders
        curr_cands (List[int], optional): Candidates to consider. Defaults to all candidates if not provided.
        p (float): Probability of choosing the winner from the Maximal Lotteries distribution
        B (float): Upper bound for elimination threshold in the Random Consensus Builder method

    Returns:
        A sorted list of candidates.
    """

    if random.random() < p:
        return [maximal_lottery.choose(profile, curr_cands=curr_cands)]

    else:
        beta = random.uniform(0.5, B)
        return random_consensus_builder_st(profile, curr_cands=curr_cands, beta=beta)
    
@vm(name="Maximal Lotteries mixed with RaDiUS")
def MLRaDiUS(profile, curr_cands=None):

    """For p, B, and the probability distribution over beta given in the proof of Theorem 5 of Charikar et al. (https://arxiv.org/abs/2306.17838), choose the winner from the Maximal Lotteries distribution with probability p; with probability 1-p, run the RaDiUS method with beta chosen according to the distribution over beta.

    Args:
        profile (Profile): An anonymous profile of linear orders
        curr_cands (List[int], optional): Candidates to consider. Defaults to all candidates if not provided.

    Returns:
        A sorted list of candidates.
    """
    # Parameters
    B = 0.876353 # given in the proof of Theorem 5 of Charikar et al. (https://arxiv.org/abs/2306.17838)

    # Calculate p as per proof
    ln3 = np.log(3)
    LB = np.log((1 + B) / (1 - B))
    I = 0.5 * (LB - ln3)  # Integral value
    p = 1 / (1 + I)

    def sample_beta(B):
        # Generate a single uniform random number
        u = np.random.uniform(0, 1)

        # Compute E(u)
        Eu = np.exp(u * (LB - ln3) + ln3)

        # Compute beta sample
        beta_sample = (Eu - 1) / (Eu + 1)

        return beta_sample
    
    if random.random() < p:
        return [maximal_lottery.choose(profile, curr_cands=curr_cands)]
    
    else:
        beta = sample_beta(B)
        return [RaDiUS.choose(profile, curr_cands=curr_cands, beta=beta)]



'''
    File: generate_utility_profiles.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: May 26, 2023
    
    Functions to generate utility profiles.
'''


from math import ceil
import numpy as np
from scipy.spatial import distance
from functools import partial

from pref_voting.utility_profiles import UtilityProfile
from pref_voting.utility_functions import *

# turn off future warnings.
# getting the following warning when calling tabulate to display a profile: 
# /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/tabulate.py:1027: FutureWarning: elementwise comparison failed; returning scalar instead, but in the future will perform elementwise comparison
#  if headers == "keys" and not rows:
# see https://stackoverflow.com/questions/40659212/futurewarning-elementwise-comparison-failed-returning-scalar-but-in-the-futur
#
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)


def generate_utility_profile_uniform(num_candidates, num_voters, num_profiles = 1):
    """
    Generate a utility profile where each voter assigns a random number between 0 and 1 to each candidate.

    Args:   
        num_candidates (int): The number of candidates.
        num_voters (int): The number of voters.
    
    Returns:
        UtilityProfile: A utility profile.

    """

    cand_utils = np.random.uniform(size=(num_profiles, num_voters, num_candidates))

    uprofs = [UtilityProfile([{c: cand_utils[pidx][v][c] 
                               for c in range(num_candidates)} 
                               for v in range(num_voters)]) 
                               for pidx in range(num_profiles)]
    
    return uprofs if num_profiles > 1 else uprofs[0]

def generate_utility_profile_normal(num_candidates, num_voters, std = 0.1, normalize = None, num_profiles = 1):
    """
    Generate a utility profile where each voter assigns a random number drawn from a normal distribution with a randomly chosen mean (between 0 and 1) with standard deviation ``std`` to each candidate.   
    
    Args:
        num_candidates (int): The number of candidates.
        num_voters (int): The number of voters. 
        std (float): The standard deviation of the normal distribution. The default is 0.1.
        normalize (str): The normalization method to use. The default is None.
    
    Returns:
        UtilityProfile: A utility profile.
    """
    
    mean_utilities = {c: np.random.uniform(0, 1) for c in range(num_candidates)}
    cand_utils = {c: np.random.normal(mean_utilities[c], std, size=(num_profiles, num_voters)) for c in range(num_candidates)}
    
    if normalize == "range": 
        uprofs = [UtilityProfile([{c: cand_utils[c][pidx][vidx] 
                                   for c in range(num_candidates)} 
                                   for vidx in range(num_voters)]).normalize_by_range() 
                                   for pidx in range(num_profiles)]
    elif normalize == "score":
        uprofs = [UtilityProfile([{c: cand_utils[c][pidx][vidx] 
                                   for c in range(num_candidates)} 
                                   for vidx in range(num_voters)]).normalize_by_standard_score() 
                                   for pidx in range(num_profiles)]
    else: # do not normalize
        uprofs = [UtilityProfile([{c: cand_utils[c][pidx][vidx] 
                                   for c in range(num_candidates)} 
                                   for vidx in range(num_voters)]) 
                                   for pidx in range(num_profiles)]


    return uprofs if num_profiles > 1 else uprofs[0]
utility_functions = {
    "RM": {
        "func": mixed_rm_utility,
        "param": 1
    },
    "Linear": {
        "func": linear_utility,
        "param": None
    },
    "Quadratic": 
    {   
        "func": quadratic_utility,
        "param": None
    },
    "Shepsle": {
        "func": shepsle_utility,
        "param": None
    },
    "City Block": { 
        "func": city_block_utility,
        "param": None
    },
    "Matthews": { 
        "func": matthews_utility,
        "param": None
    }

}
def generate_spatial_utility_profile(num_cands, 
                                     num_voters, 
                                     num_dims = 2, 
                                     utility_function = "Quadratic", 
                                     utility_function_param = None):
    
    """
    Create a spatial utility profile using specified utility functions. 


    Args:
        num_cands (int): The number of candidates.
        num_voters (int): The number of voters.
        num_dims (int): The number of dimensions. The default is 2.
        utility_function (str): The utility function to use. The default is "Linear".
        utility_function_param (float): The parameter of the utility function. The default is None.
        
    Returns:
        UtilityProfile: A spatial utility profile.
    """

    # the first component of the parameter is the number of dimensions, 
    # the second component is used to define the mixed model: 
    # beta = 1 is proximity model (i.e., squared Euclidean distance)

    mean = [0] * num_dims  # mean is 0 for each dimension
    cov = np.diag([1] * num_dims)  # diagonal covariance

    _utility_fnc = utility_functions[utility_function]["func"]

    if utility_functions[utility_function]["param"] is not None or utility_function_param is not None: 
        util_parm = utility_function_param  if utility_function_param is not None else utility_functions[utility_function]["param"]
        utility_fnc = partial(_utility_fnc, util_parm)

    else: 
        utility_fnc = _utility_fnc

    # sample candidate/voter positions using a multivariate normal distribution
    cand_positions = np.random.multivariate_normal(np.array(mean), cov, num_cands)
    voter_positions = np.random.multivariate_normal(np.array(mean), cov, num_voters)

    utilities = [{c: utility_fnc(v_pos, c_pos)  for c, c_pos in enumerate(cand_positions)} 
                 for _, v_pos in enumerate(voter_positions)]

    return UtilityProfile(utilities)





'''
    File: create_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: August 8, 2024
    
'''

def compose(vm1, vm2):
    """After restricting the profile to the set of vm1 winners, run vm2

    Args:
        vm1, vm2 (VotingMethod): The voting methods to be composed.

    Returns:
        A VotingMethod that composes vm1 and vm2.

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import compose
        from pref_voting.scoring_methods import borda
        from pref_voting.c1_methods import copeland

        prof = Profile([[1, 3, 0, 2], [2, 1, 3, 0], [3, 0, 2, 1]], [1, 2, 1])

        prof.display()

        copeland_borda = compose(copeland, borda)

        copeland.display(prof)
        borda.display(prof)
        copeland_borda.display(prof)

    """

    def _vm(edata, curr_cands=None):

        vm1_ws = vm1(edata, curr_cands=curr_cands)

        return vm2(edata, curr_cands=vm1_ws)

    return VotingMethod(_vm, name=f"{vm1.name}-{vm2.name}")


def _compose(vm1, vm2):
    """
    Same as compose, but used to make it easier to document composed voting methods.
    """

    def _vm(edata, curr_cands=None):

        vm1_ws = vm1(edata, curr_cands=curr_cands)

        return vm2(edata, curr_cands=vm1_ws)

    return _vm

def faceoff(vm1, vm2):
    """If the vm1 and vm2 winners are the same, return that set of winners. Otherwise, for each choice of a vm1 winner A and vm2 winner B, add to the ultimate winners whichever of A or B is majority preferred to the other (or both if they are tied).

    Args:
        vm1, vm2 (VotingMethod): The voting methods to faceoff.

    Returns:
        A VotingMethod that runs the faceoff of vm1 and vm2.

    """

    def _vm(edata, curr_cands=None):

        curr_cands = edata.candidates if curr_cands is None else curr_cands

        vm1_winners = vm1(edata, curr_cands)
        vm2_winners = vm2(edata, curr_cands)

        if vm1_winners == vm2_winners:
            return vm1_winners
        
        else:
            winners = list()

            for a in vm1_winners:
                for b in vm2_winners:
                    if edata.margin(a,b) > 0:
                        winners.append(a)
                    elif edata.margin(b,a) > 0:
                        winners.append(b)
                    elif edata.margin(a,b) == 0:
                        winners.append(a)
                        winners.append(b) 

            return list(set(winners))

    return VotingMethod(_vm, name=f"{vm1.name}-{vm2.name} Faceoff")

def _faceoff(vm1, vm2):
    """
    Same as faceoff, but used to make it easier to document faceoff voting methods.
    """

    def _vm(edata, curr_cands=None):

        curr_cands = edata.candidates if curr_cands is None else curr_cands

        vm1_winners = vm1(edata, curr_cands)
        vm2_winners = vm2(edata, curr_cands)

        if vm1_winners == vm2_winners:
            return vm1_winners
        
        else:
            winners = list()

            for a in vm1_winners:
                for b in vm2_winners:
                    if edata.margin(a,b) > 0:
                        winners.append(a)
                    elif edata.margin(b,a) > 0:
                        winners.append(b)
                    elif edata.margin(a,b) == 0:
                        winners.append(a)
                        winners.append(b) 

            return list(set(winners))

    return _vm